#!/usr/bin/env python3

import shutil
from pathlib import Path


def main():
    """This script is used to update the ref.*.out files in the test/references directory
    with the result.*.out files in the test/results directory.

    This script is used when the results of the program are changed.
    """
    this_file_dir = Path(__file__).resolve().parent
    results_dir = this_file_dir / "results"
    references_dir = this_file_dir / "references"
    result_list = list(results_dir.glob("result.*.out"))
    for result in result_list:
        # if result includes "multi-process", skip it
        if "multi-process" in str(result):
            continue
        # replace only the first occurrence of "result" with "reference"
        ref_name = result.name.replace("result", "ref", 1)
        ref_path = references_dir / ref_name
        shutil.copy2(result, ref_path)


if __name__ == "__main__":
    main()



from pydantic import BaseModel

from sum_dirac_dfcoef.functions_info import FunctionsInfo
from sum_dirac_dfcoef.utils import is_float, space_separated_parsing


class Coefficient(BaseModel, validate_assignment=True):
    """This class is used to store the specific coefficient information.

    e.g.  35  L Ag U  dzz     0.0000000014  0.0000000000  0.0000000000  0.0000000000
    -> Coefficient(
        vector_num=35,
        symmetry_label="Ag",
        atom_azimuthal="Ud",
        magnetic_label="zz",
        need_identifier=False,
        coefficient=1.96E-18,
        idx_within_same_atom=0,
        multiplication=1
    )
    Attributes:
        vector_num (int): Serial number of the vector
        symmetry_label (str): The symmetry label to identify the vector
        atom_label (str): The atom label to identify the vector
        azimuthal_label (str): The azimuthal quantum number label to identify the vector
        magnetic_label (str): The magnetic quantum number label to identify the vector
        need_identifier (bool): Whether the idx_within_same_atom is needed to print to the output file or not.
        coefficient (float): Coefficient value
        idx_within_same_atom (int): Index of the order of atoms in the same AtomLabel
        multiplication (int): The multiplicity of the specified atom label.
    """

    vector_num: int
    symmetry_label: str
    atom_label: str
    azimuthal_label: str
    magnetic_label: str
    need_identifier: bool
    coefficient: float
    idx_within_same_atom: int
    multiplication: int

    def __repr__(self) -> str:
        super().__repr__()
        return f"Coefficient(vector_num: {self.vector_num}, \
symmetry_label: {self.symmetry_label}, \
atom_label: {self.atom_label}, \
azimuthal_label: {self.azimuthal_label}, \
magnetic_label: {self.magnetic_label}, \
need_identifier: {self.need_identifier}, \
coefficient: {self.coefficient}, \
idx_within_same_atom: {self.idx_within_same_atom}, \
multiplication: {self.multiplication})"


def get_coefficient(line_str: str, orbitals: FunctionsInfo, idx_within_same_atom: int) -> Coefficient:
    """
    This function parses the line that contains the coefficient and returns the Coefficient object.
    (e.g.)
    sym_and_atom_and_orb_str = "B3gCldyz"
    symmetry_type = "B3g"
    atom_type = "Cl"
    orbital_type = "dyz"
    """
    # ref (print ): https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/dirac/dirout.F#L388-389
    # ref (format): https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/dirac/dirout.F#L453
    # FORMAT(3X,I5,2X,A12,2X,4F14.10)
    # https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/dirac/dirtra.F#L168-169

    def parse_line() -> Coefficient:
        """
        This function parses the line that contains the coefficient.

        line write source: https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/dirac/dirout.F#L440-442
        line format source: https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/dirac/dirout.F#L453
        line format : FORMAT(3X,I5,2X,A12,2X,4F14.10)
        """
        words = space_separated_parsing(line_str)
        # JS (I5): Serial number of the vector
        vec_num = int(words[0])

        #
        # PLABEL(IPLAB(IBAS(IFRP)+JS,2),2) (A12): Information about the vector to identify the vector
        #
        # PLABEL source: https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/dirac/dirtra.F#L168-169
        # PLABEL(NLAB,2) = CLS(IC)//' '//REP(IRP)//NAMN(ICENT)(1:3)//GTOTYP(ITYP)
        # CLS (A1): https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/dirac/dirtra.F#L45
        # REP (A3): https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/include/pgroup.h#L16
        # NAMN (A3, defined as A4, but only (1:3) is used): https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/include/nuclei.h#L25
        # GTOTYP (A4): https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/include/ccom.h#L8
        component_func = "large" if line_str[10] == "L" else ("small" if line_str[10] == "S" else "")  # CLS
        symmetry_label = line_str[12:15].strip().replace(" ", "")  # REP (e.g. "Ag ")
        atom_label = line_str[15:18].strip().replace(" ", "")  # NAMN (e.g. "Ag U  dzz " => "U")
        azimuthal_label = line_str[18:19].strip().replace(" ", "")  # NAMN (e.g. "Ag U  dzz   " => "d")
        magnetic_label = line_str[19:22].strip().replace(" ", "")  # GTOTYP (e.g. "Ag U  dzz " => "zz")

        # COEF (4F14.10)
        # coefficients = [line_str[24:38], line_str[38:52], line_str[52:66], line_str[66:80]]
        coef_num = 4
        coef_len = 14
        coef_start_idx = 24
        coefficient = sum(
            pow(float(line_str[i : i + coef_len]) if is_float(line_str[i : i + coef_len]) else -100, 2)
            for i in range(coef_start_idx, coef_start_idx + coef_len * coef_num, coef_len)
        )

        need_identifier = (
            True if len(orbitals[component_func][symmetry_label][atom_label]) > 1 or orbitals[component_func][symmetry_label][atom_label][idx_within_same_atom].mul > 1 else False
        )
        multiplication = int(orbitals[component_func][symmetry_label][atom_label][idx_within_same_atom].mul)

        return Coefficient(
            vector_num=vec_num,
            symmetry_label=symmetry_label,
            atom_label=atom_label,
            azimuthal_label=azimuthal_label,
            magnetic_label=magnetic_label,
            need_identifier=need_identifier,
            coefficient=coefficient,
            idx_within_same_atom=idx_within_same_atom,
            multiplication=multiplication,
        )

    """
    Main function to get coefficient
    """
    coef = parse_line()

    return coef



from pathlib import Path
from typing import List

from sum_dirac_dfcoef.args import args
from sum_dirac_dfcoef.data import DataMO
from sum_dirac_dfcoef.header_info import HeaderInfo
from sum_dirac_dfcoef.utils import debug_print


class OutputFileWriter:
    """This class has methods to write data to the output file."""

    def __init__(self) -> None:
        super().__init__()
        self.output_path = self.get_output_path()

    def write_no_header_info(self) -> None:
        # Print NO_HEADERINFO twice because the first and second lines are used for the header
        # If user uses --no-scf option, we cannot get the eigenvalues and the number of electrons from the output file.
        # Therefore, print NO_HEADERINFO twice to avoid using the output file for dcaspt2_input_generator program.
        with open(self.output_path, "a", encoding="utf-8") as f:
            msg = "NO_HEADERINFO: This output cannot be used for the dcaspt2_input_generator program.\n"
            f.write(msg)
            f.write(msg)

    def write_headerinfo(self, header_info: HeaderInfo) -> None:
        with open(self.output_path, "a", encoding="utf-8") as f:
            scheme = "default" if header_info.scheme.value == 0 else str(header_info.scheme.value)
            line = f"electron_num {header_info.electrons} point_group {header_info.point_group} moltra_scheme {scheme}\n"
            for symmetry_type, d in header_info.moltra_info.range_dict.items():
                line += f"{symmetry_type} {d} "
            line += "\n"
            for symmetry_type, d in header_info.eigenvalues.shell_num.items():
                line += f"{symmetry_type} "
                for eigenvalue_type, num in d.items():
                    # only write closed, open, virtual (positive energy eigenvalues)
                    if eigenvalue_type in ("closed", "open", "virtual"):
                        line += f"{eigenvalue_type} {num} "
            line += "\n"
            f.write(line)

    def write_mo_data(self, mo_data: List[DataMO]) -> None:
        with open(self.output_path, "a", encoding="utf-8") as f:
            f.write("\n")
            for mo in mo_data:
                digit_int = len(str(int(mo.mo_energy)))  # number of digits of integer part
                # File write but if args.compress is True \n is not added
                mo_info_energy = f"{mo.mo_info} {mo.mo_energy:{digit_int}.{args.decimal}f}" + ("\n" if not args.compress else "")
                f.write(mo_info_energy)

                for c in mo.coef_list:
                    for idx in range(c.multiplication):
                        percentage = c.coefficient / mo.norm_const_sum * 100
                        sym_label = c.symmetry_label if not args.ignore_sym else ""
                        ml_label = c.magnetic_label if not args.ignore_ml else ""
                        func_label = f"{sym_label}{c.atom_label}{c.azimuthal_label}{ml_label}"
                        atomic_symmetry_label = f"{func_label}({c.idx_within_same_atom + idx})" if c.need_identifier else func_label
                        output_str: str
                        if args.compress:
                            output_str = f" {atomic_symmetry_label} {percentage:.{args.decimal}f}"
                        else:
                            output_str = f"{atomic_symmetry_label:<12} {percentage:{args.decimal+4}.{args.decimal}f} %\n"
                        f.write(output_str)
                f.write("\n")  # add empty line
                debug_print(f"sum of coefficient {mo.norm_const_sum:.{args.decimal}f}")

    def create_blank_file(self) -> None:
        # Open the file in write mode
        # Even if the file already exists, it will be overwritten with a blank file
        file = open(self.output_path, "w", encoding="utf-8")
        file.close()

    def get_output_path(self) -> Path:
        import sys

        if args.output is None:
            output_name = "sum_dirac_dfcoef.out"
            output_path = Path.absolute(Path.cwd() / output_name)
        else:
            output_path = Path(args.output).expanduser().resolve()
            if output_path.is_dir():
                sys.exit(
                    "ERROR: The path you specified as the sum_dirac_dfcoef output is a directory. Not a file.\
Please check your -o or --output option is correct."
                )

        return output_path


output_file_writer = OutputFileWriter()



#!/usr/bin/env python3
from sum_dirac_dfcoef.args import args
from sum_dirac_dfcoef.file_writer import output_file_writer
from sum_dirac_dfcoef.functions_info import get_functions_info
from sum_dirac_dfcoef.header_info import HeaderInfo
from sum_dirac_dfcoef.privec_reader import PrivecProcessor
from sum_dirac_dfcoef.utils import get_dirac_filepath, should_write_electronic_results_to_file, should_write_positronic_results_to_file


def main() -> None:
    dirac_filepath = get_dirac_filepath()
    dirac_output = open(dirac_filepath, encoding="utf-8")
    dirac_output.seek(0)  # rewind to the beginning of the file
    header_info = HeaderInfo()
    if args.for_generator:
        header_info.read_header_info(dirac_output)
    dirac_output.seek(0)
    functions_info = get_functions_info(dirac_output)
    output_file_writer.create_blank_file()

    dirac_output_lines = dirac_output.readlines()
    # Read coefficients from the output file of DIRAC and store them in data_all_mo.
    privec_processor = PrivecProcessor(dirac_output_lines, functions_info, header_info.eigenvalues)
    privec_processor.read_privec_data_wrapper()

    # Write the header information to the output file.
    if args.for_generator:
        header_info.calculate_moltra_idx_range(privec_processor.data_all_mo)
        output_file_writer.write_headerinfo(header_info)
    else:
        # If the output file is not for dcaspt2_input_generator, don't write header information.
        output_file_writer.write_no_header_info()

    # Sort the MOs by energy.
    if not args.no_sort:
        privec_processor.data_all_mo.sort_mo_energy()

    # Write the MO data to the output file.
    if should_write_electronic_results_to_file():
        output_file_writer.write_mo_data(privec_processor.data_all_mo.electronic)
    if should_write_positronic_results_to_file():
        output_file_writer.write_mo_data(privec_processor.data_all_mo.positronic)



import argparse
import os
import sys


class PrintVersionExitAction(argparse.Action):
    """Print version and exit if -v or --version option is used."""

    def __init__(self, option_strings, dest=argparse.SUPPRESS, default=argparse.SUPPRESS, help=None):  # noqa: A002
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):  # noqa: ARG002
        from sum_dirac_dfcoef.__about__ import __version__

        print(f"{__version__}")
        sys.exit()


class PrintHelpArgumentParser(argparse.ArgumentParser):
    """Print help message and exit if error occurs during parsing arguments."""

    def error(self, message):
        self.print_help(sys.stdout)
        err_msg = f"{self.prog}: error: {message}\n"
        self.exit(2, err_msg)


def parse_args() -> "argparse.Namespace":
    parser = PrintHelpArgumentParser(
        description="Summarize the coefficients from DIRAC output file that *PRIVEC option is used. (c.f. http://www.diracprogram.org/doc/master/manual/analyze/privec.html)"
    )
    parser.add_argument(
        "-i", "--input", type=str, required=True, help="(required) file path of DIRAC output. Please quote if the path include spaces.", dest="input", metavar='"INPUT"'
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="File path of sum_dirac_dfcoef output. Default: sum_dirac_dfcoef.out. Please quote if the path include spaces.",
        dest="output",
        metavar='"OUTPUT"',
    )
    parser.add_argument(
        "-g",
        "--for-generator",
        action="store_true",
        help="Automatically set the arguments for dcaspt2_input_generator. \
This option is useful when you want to use the result of this program as input to dcaspt2_input_generator. \
This option is equivalent to set -c/--compress and not set -p/--positronic and --no-scf options.",
        dest="for_generator",
    )
    parser.add_argument(
        "-j",
        "--parallel",
        type=int,
        nargs="?",
        const=-1,
        default=1,
        help="Number of parallel processes. Default: 1 (single process).\
        If you set -j option without argument, the number of parallel processes is set to the number of CPU cores(=os.cpu_count()).",
        dest="parallel",
    )
    parser.add_argument(
        "-c",
        "--compress",
        action="store_true",
        help="Compress output. Display coefficients on one line for each kramers pair.\
This options is useful when you want to use the result in a spreadsheet like Microsoft Excel.",
        dest="compress",
    )
    parser.add_argument(
        "-t",
        "--threshold",
        type=float,
        default=0.1,
        help="threshold. Default: 0.1 %% (e.g) --threshold=0.1 → only print atomic orbitals with more than 0.1 %% contribution",
        dest="threshold",
    )
    parser.add_argument(
        "-d",
        "--decimal",
        type=int,
        default=5,
        choices=range(1, 16),
        help="Set the decimal places. Default: 5 (e.g) --decimal=3 → print orbital with 3 decimal places (0.123, 2.456, ...). range: 1-15",
        dest="decimal",
    )
    parser.add_argument(
        "--ignore-ml",
        action="store_true",
        help="Ignore the magnetic quantum number label. This option is useful when you want to sum the coefficients of the same atomic orbital\
            except for the magnetic quantum number label.",
        dest="ignore_ml",
    )
    parser.add_argument(
        "--ignore-sym",
        action="store_true",
        help="Ignore symmetry label (e.g. Ag, B3g, ...). This option is useful when you want to sum the coefficients of the same atomic orbital\
            except for the symmetry label.",
        dest="ignore_sym",
    )
    parser.add_argument("-a", "--all-write", action="store_true", help="Print all kramers pairs(Positronic and Electronic).", dest="all_write")
    parser.add_argument(
        "-p",
        "--positronic-write",
        action="store_true",
        help="Print only Positronic kramers pairs. The output with this option cannot be used as input to dcaspt2_input_generator.",
        dest="positronic_write",
    )
    parser.add_argument("-v", "--version", action=PrintVersionExitAction, help="Print version and exit", dest="version")
    parser.add_argument(
        "--no-scf",
        action="store_true",
        help="If you don't activate .SCF keyword in your DIRAC input file, you must use this option.\
            But you cannot use the output using this option to dcaspt2_input_generator program.",
        dest="no_scf",
    )
    parser.add_argument("--debug", action="store_true", help="print debug output (Normalization constant, Sum of kramers pair coefficient)", dest="debug")
    parser.add_argument("--no-sort", action="store_true", help="Don't sort the output by kramers pair energy")
    # If -v or --version option is used, print version and exit
    args = parser.parse_args()

    args.parallel = os.cpu_count() if args.parallel == -1 else args.parallel

    # -g and -p, --no-scf options are exclusive
    if args.for_generator and args.positronic_write:
        msg = "-g/--for-generator and -p/--positronic-write options cannot be set at the same time \
because dcaspt2_input_generator needs data of electronic orbitals."
        parser.error(msg)
    if args.for_generator and args.no_scf:
        msg = "-g/--for-generator and --no-scf options cannot be set at the same time \
because dcaspt2_input_generator needs eigenvalues information after SCF calculation."
        parser.error(msg)

    if args.for_generator:
        args.no_scf = False
        args.compress = True
        args.positronic_write = False

    if not (args.no_scf or args.positronic_write) and args.compress:
        args.for_generator = True

    if args.all_write and args.positronic_write:
        parser.error("-a/--all-write and -p/--positronic-write options cannot be set at the same time.")

    return args


args = parse_args()



import pickle
import re
import sys
from pathlib import Path
from typing import Any, List


def space_separated_parsing(line: str) -> List[str]:
    return [word for word in line.rstrip("\n").split(" ") if word != ""]


def space_separated_parsing_upper(line: str) -> List[str]:
    return [word.upper() for word in line.rstrip("\n").split(" ") if word != ""]


def fast_deepcopy_pickle(obj: Any) -> Any:
    """Fast deep copy using pickle.

    Args:
        obj (Any): object to be copied

    Returns:
        Any: copied object
    """
    return pickle.loads(pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL))  # noqa: S301


def debug_print(message: str) -> None:
    from sum_dirac_dfcoef.args import args

    # print debug message if --debug option is used
    if args.debug:
        print(message)


def is_float(parameter: str) -> bool:
    try:
        float(parameter)
        return True
    except ValueError:
        return False


def is_start_dirac_input_field(line: str) -> bool:
    return True if "Contents of the input file" in line else False


def is_end_dirac_input_field(line: str) -> bool:
    return True if "Contents of the molecule file" in line else False


def is_dirac_input_keyword(word: str) -> bool:
    regex_keyword = r" *\.[0-9A-Z]+"
    return re.match(regex_keyword, word) is not None


def is_dirac_input_section(word: str) -> bool:
    regex_section = r" *\*{1,2}[0-9A-Z]+"
    return re.match(regex_section, word) is not None


def is_dirac_input_section_one_star(word: str) -> bool:
    regex_section = r" *\*[0-9A-Z]+"
    return re.match(regex_section, word) is not None


def is_dirac_input_section_two_stars(word: str) -> bool:
    regex_section = r" *\*{2}[0-9A-Z]+"
    return re.match(regex_section, word) is not None


def is_dirac_input_line_comment_out(word: str) -> bool:
    regex_comment_out = r" *[!#]"
    return re.match(regex_comment_out, word) is not None


def is_dirac_input_line_should_be_skipped(words: List[str]) -> bool:
    if len(words) == 0:
        return True
    if is_dirac_input_line_comment_out(words[0]):
        return True
    return False


def delete_dirac_input_comment_out(line: str) -> str:
    regex_comment_out = r" *[!#]"
    idx_comment_out = re.search(regex_comment_out, line)
    if idx_comment_out is None:
        return line
    return line[: idx_comment_out.start()]


def get_dirac_filepath() -> Path:
    from sum_dirac_dfcoef.args import args

    input_path = Path(args.input).expanduser().resolve()
    if not input_path.exists():
        sys.exit(f"ERROR: DIRAC output file is not found. file={input_path}")
    elif input_path.is_dir():
        sys.exit(
            "ERROR: The path you specified as the DIRAC output is a directory. Not a file.\
Please check your -i or --input option is correct."
        )
    return input_path


def should_write_positronic_results_to_file() -> bool:
    from sum_dirac_dfcoef.args import args

    if args.all_write or args.positronic_write:
        return True
    else:
        return False


def should_write_electronic_results_to_file() -> bool:
    from sum_dirac_dfcoef.args import args

    if args.all_write or not args.positronic_write:
        return True
    else:
        return False



from io import TextIOWrapper
from typing import ClassVar, Dict, List

from sum_dirac_dfcoef.utils import (
    delete_dirac_input_comment_out,
    is_dirac_input_keyword,
    is_dirac_input_line_should_be_skipped,
    is_dirac_input_section,
    is_dirac_input_section_two_stars,
    is_end_dirac_input_field,
    is_start_dirac_input_field,
    space_separated_parsing_upper,
)


class MoltraInfo:
    """Class to store MOLTRA information for the sum_dirac_dfcoef module.

    Attributes:
        is_default (bool): True if the .ACTIVE section is not found
        range_str (List[str]): List of strings in the .ACTIVE section
        range_dict (Dict[str, str]): Dictionary of the .ACTIVE section. Key: symmetry type, Value: range string
    """

    is_default: bool = True
    range_str: ClassVar[List[str]] = []
    # range_str Example:
    # ['energy -20 10 2', '10..180', ...]
    range_dict: ClassVar[Dict[str, str]] = {}

    @classmethod
    def read_moltra_section(cls, dirac_output: TextIOWrapper):
        """Read the MOLTRA section settings from the output file of DIRAC

        Args:
            dirac_output (TextIOWrapper): Output file of DIRAC

        Returns:
            None (cls.range_str and cls.is_default will be updated)
        """

        is_moltra_section = False
        is_reach_input_field = False
        is_next_line_active = False
        for line in dirac_output:
            no_comment_line = delete_dirac_input_comment_out(line)
            words = space_separated_parsing_upper(no_comment_line)
            if is_dirac_input_line_should_be_skipped(words):
                continue

            if is_start_dirac_input_field(line):
                is_reach_input_field = True
                continue

            if is_end_dirac_input_field(line):
                break  # end of input field

            if is_reach_input_field:
                if is_dirac_input_section_two_stars(words[0]):
                    if "**MOLTRA" in words[0]:
                        is_moltra_section = True
                        continue

                if is_moltra_section:
                    if ".ACTIVE" in words[0]:
                        cls.is_default = False
                        is_next_line_active = True
                        continue

                if is_next_line_active:
                    if is_dirac_input_section(words[0]) or is_dirac_input_keyword(words[0]):
                        # End of the .ACTIVE section
                        break
                    cls.range_str.append(no_comment_line.strip())



from io import TextIOWrapper

from sum_dirac_dfcoef.utils import (
    delete_dirac_input_comment_out,
    is_dirac_input_keyword,
    is_dirac_input_line_should_be_skipped,
    is_dirac_input_section,
    is_dirac_input_section_two_stars,
    is_end_dirac_input_field,
    is_start_dirac_input_field,
    space_separated_parsing_upper,
)


class Scheme:
    """Class to store **MOLTRA > .SCHEME information for the sum_dirac_dfcoef module.

    Attributes:
        value (int): 0 if .SCHEME was not set, otherwise equals to the value of .SCHEME option.
    """

    def __init__(self) -> None:
        self.value: int = 0

    def get_scheme_num_from_input(self, dirac_output: TextIOWrapper) -> None:
        """If user explicitly set **MOLTRA > .SCHEME option (https://diracprogram.org/doc/release-23/manual/moltra.html#scheme),
        read the option value and store it to self.value.
        If this option is not specified, self.value is kept at 0 (means scheme was not set).

        Args:
            dirac_output (TextIOWrapper): Output file of DIRAC

        Returns:
            None (self.value will be updated)
        """
        is_reach_input_field = False
        is_moltra_section = False
        is_next_line_scheme = False
        for line in dirac_output:
            no_comment_out_line = delete_dirac_input_comment_out(line)
            words = space_separated_parsing_upper(no_comment_out_line)

            if is_dirac_input_line_should_be_skipped(words):
                continue

            if is_start_dirac_input_field(no_comment_out_line):
                is_reach_input_field = True
                continue

            if is_end_dirac_input_field(no_comment_out_line):
                break  # end of input field

            if is_reach_input_field:
                if is_dirac_input_section_two_stars(words[0]):
                    if "**MOLTRA" in words[0]:
                        is_moltra_section = True
                        continue

                if is_moltra_section:
                    if ".SCHEME" in words[0]:
                        is_next_line_scheme = True
                        continue

                if is_next_line_scheme:
                    if is_dirac_input_section(words[0]) or is_dirac_input_keyword(words[0]):
                        # End of the .ACTIVE section
                        break
                    self.value = int(words[0])
                    if self.value <= 0:
                        msg = f"Invalid **MOLTRA > .SCHEME value. This option must be larger than 1, but actual .SCHEME value is {self.value}."
                        raise ValueError(msg)



import concurrent.futures
from enum import Enum, auto
from typing import Dict, List, Tuple

from sum_dirac_dfcoef.args import args
from sum_dirac_dfcoef.atoms import AtomInfo
from sum_dirac_dfcoef.coefficient import get_coefficient
from sum_dirac_dfcoef.data import DataAllMO, DataMO
from sum_dirac_dfcoef.eigenvalues import Eigenvalues
from sum_dirac_dfcoef.functions_info import FunctionsInfo
from sum_dirac_dfcoef.utils import debug_print, fast_deepcopy_pickle, space_separated_parsing


class STAGE(Enum):
    #                                                                             SKIP_READING_COEF
    #                                                                                    ↑↓
    # STAGE TRANSITION: INIT -> SKIP_AFTER_VECTOR_PRINT_LINE -> VECTOR_PRINT -> WAIT_END_READING_COEF -> END
    #                                                                             ↓               ↑
    #                                                                      WAIT_FIRST_COEF -> READING_COEF
    INIT = auto()
    SKIP_AFTER_VECTOR_PRINT_LINE = auto()
    VECTOR_PRINT = auto()
    WAIT_END_READING_COEF = auto()
    SKIP_READING_COEF = auto()
    WAIT_FIRST_COEF = auto()
    READING_COEF = auto()
    END = auto()


class PrivecProcessor:
    """This class has methods to read coefficients from the output file of DIRAC and store them in self.data_all_mo (final result).

    Attributes:
        dirac_output (List[str]): Output file strings of DIRAC
        stage (STAGE): Stage of reading coefficients
        is_electronic (bool): True if the current MO is electronic
        is_less_than_dirac_21 (bool): True if the current version of DIRAC is less than 21.
                                      Use this variable to determine which attribute (func_idx_dirac21/19) of AtomInfo should be used.
        eigenvalues (Eigenvalues): Eigenvalues
        mo_sym_type (str): Symmetry type of the current MO
        functions_info (FunctionsInfo): FunctionsInfo (used to handle which functions are used in the current MO)
        data_mo (DataMO): DataMO (temporary result of current reading MO)
        data_all_mo (DataAllMO): DataAllMO (final result)
        used_atom_info (Dict[str, AtomInfo]): Used AtomInfo
        current_atom_info (AtomInfo): Current AtomInfo
    """

    def __init__(self, dirac_output: List[str], functions_info: FunctionsInfo, eigenvalues: Eigenvalues) -> None:
        self.dirac_output = dirac_output
        self.stage = STAGE.INIT
        self.is_electronic = False
        self.is_less_than_dirac_21 = False
        self.eigenvalues = eigenvalues
        self.mo_sym_type = ""
        self.functions_info = functions_info
        self.data_mo = DataMO()
        self.data_all_mo = DataAllMO()
        self.used_atom_info: Dict[str, AtomInfo] = {}
        self.current_atom_info = AtomInfo()

    def read_privec_data(self, rank: int = 0) -> Tuple[DataAllMO, Eigenvalues]:
        """Read coefficients from the output file of DIRAC and store them in data_all_mo.

        self.data_all is the final result of this function. You can get all results from this variable except header information.
        """
        self.stage = STAGE.INIT
        mo_cnt = 0
        for line_str in self.dirac_output:
            if self.stage == STAGE.END:
                break  # End of reading coefficients
            elif self.stage == STAGE.SKIP_READING_COEF:
                if len(line_str.strip()) == 0:
                    self.transition_stage(STAGE.WAIT_END_READING_COEF)
                continue
            elif self.stage == STAGE.SKIP_AFTER_VECTOR_PRINT_LINE:
                self.transition_stage(STAGE.VECTOR_PRINT)
                continue

            words = space_separated_parsing(line_str)

            if self.need_to_skip_this_line(words):
                if self.stage == STAGE.VECTOR_PRINT and self.detect_next_titler(line_str):
                    msg = "WARNING: The next title is detected before the end of reading coefficients.\n\
In order to force DIRAC to print vector print, please add .ANALYZE and .PRIVEC option to the input file of DIRAC."
                    print(msg)
                    self.transition_stage(STAGE.END)
                elif self.stage == STAGE.READING_COEF:
                    if self.need_to_create_results_for_current_mo(words):
                        self.add_current_mo_data_to_data_all_mo()
                        self.transition_stage(STAGE.WAIT_END_READING_COEF)

            elif self.stage == STAGE.INIT:
                if self.check_start_vector_print(words):
                    self.transition_stage(STAGE.SKIP_AFTER_VECTOR_PRINT_LINE)

            elif self.stage == STAGE.VECTOR_PRINT:
                if self.need_to_get_mo_sym_type(words):
                    self.mo_sym_type = words[2]
                    self.transition_stage(STAGE.WAIT_END_READING_COEF)

            elif self.stage == STAGE.WAIT_END_READING_COEF:
                if self.need_to_get_mo_sym_type(words):
                    self.mo_sym_type = words[2]
                elif self.need_to_start_mo_section(words):
                    self.start_mo_section(words)
                    self.transition_stage(STAGE.WAIT_FIRST_COEF)
                elif self.check_end_vector_print(words):
                    self.transition_stage(STAGE.END)

            elif self.stage == STAGE.WAIT_FIRST_COEF:
                if self.is_this_row_for_coefficients(words):
                    if args.parallel == 1 or mo_cnt % args.parallel == rank:
                        # Need to read coefficients of the current MO
                        self.add_coefficient(line_str)
                        self.transition_stage(STAGE.READING_COEF)
                    else:
                        # Don't need to read coefficients of the current MO because it is read by another process.
                        # multi-process version only
                        self.transition_stage(STAGE.SKIP_READING_COEF)
                    mo_cnt += 1

            elif self.stage == STAGE.READING_COEF:
                if self.is_this_row_for_coefficients(words):
                    self.add_coefficient(line_str)

        return self.data_all_mo, self.eigenvalues

    def transition_stage(self, new_stage: STAGE) -> None:
        self.stage = new_stage

    def is_this_row_for_coefficients(self, words: List[str]) -> bool:
        # min: 4 coefficients and other words => 5 words
        return True if 5 <= len(words) <= 9 and words[0].isdigit() else False

    def need_to_skip_this_line(self, words: List[str]) -> bool:
        return True if len(words) <= 1 else False

    def need_to_create_results_for_current_mo(self, words: List[str]) -> bool:
        return True if self.stage == STAGE.READING_COEF and len(words) <= 1 else False

    def detect_next_titler(self, line_str: str) -> bool:
        """Detect all characters are asterisk or space or line break."""
        stripped_line = line_str.strip().strip("\r\n")
        if len(stripped_line) == 0:
            return False
        return all(c == "*" for c in stripped_line)

    def need_to_get_mo_sym_type(self, words: List[str]) -> bool:
        return True if len(words) == 3 and words[0] == "Fermion" and words[1] == "ircop" else False

    def need_to_start_mo_section(self, words: List[str]) -> bool:
        if words[1] in ("Electronic", "Positronic") and words[2] == "eigenvalue" and "no." in words[3]:
            return True
        return False

    def check_start_vector_print(self, words: List[str]) -> bool:
        # ****************************** Vector print ******************************
        if len(words) < 4:
            return False
        elif words[1] == "Vector" and words[2] == "print":
            return True
        return False

    def check_end_vector_print(self, words: List[str]) -> bool:
        # https://github.com/RQC-HU/sum_dirac_dfcoef/issues/7#issuecomment-1377969626
        if len(words) >= 2 and self.stage == STAGE.WAIT_END_READING_COEF:
            return True
        return False

    def get_mo_info(self, eigenvalue_no: int) -> str:
        if args.compress:
            return f"{self.mo_sym_type} {eigenvalue_no}"
        elif self.is_electronic:
            return f"Electronic no. {eigenvalue_no} {self.mo_sym_type}"
        else:
            return f"Positronic no. {eigenvalue_no} {self.mo_sym_type}"

    def start_mo_section(self, words: List[str]) -> None:
        """
        (e.g.)
        words = ["*", "Electronic", "eigenvalue", "no.", "22:", "-2.8417809384721"]
        words = ["*", "Electronic", "eigenvalue", "no.122:", "-2.8417809384721"]
        """

        def set_is_electronic() -> None:
            if words[1] == "Positronic":
                self.is_electronic = False
            elif words[1] == "Electronic":
                self.is_electronic = True
            else:
                msg = f"ERROR: UnKnow MO type, MO_Type={words[1]}"
                raise ValueError(msg)

        def get_eigenvalue_no():
            try:
                eigenvalue_no = int(words[-2][:-1].replace("no.", ""))
            except ValueError:
                # If *** is printed, we have no information about what number this MO is.
                # Therefore, we assume that eigenvalue_no is the next number after prev_eigenvalue_no.
                prev_eigenvalue_no = self.data_mo.eigenvalue_no  # prev_electron is the number of electrons of the previous MO
                eigenvalue_no = prev_eigenvalue_no + 1
            return eigenvalue_no

        set_is_electronic()
        eigenvalue_no = get_eigenvalue_no()
        mo_energy = float(words[-1])
        mo_info = self.get_mo_info(eigenvalue_no)

        # Here is the start point of reading coefficients of the current MO
        self.data_mo.reset()  # reset data_mo because we need to delete data_mo of the previous MO
        self.data_mo.eigenvalue_no = eigenvalue_no
        self.data_mo.mo_energy = mo_energy
        self.data_mo.sym_type = self.mo_sym_type
        self.data_mo.mo_info = mo_info
        self.used_atom_info.clear()  # reset used_atom_info because we need to delete used_atom_info of the previous MO
        self.current_atom_info = AtomInfo()  # reset current_atom_info because self.current_atom_info.count_remaining_functions() may be larger than 0

    def add_coefficient(self, line_str: str) -> None:
        def get_func_no() -> int:
            try:
                num_functions = int(line_str[:10])
            except ValueError as e:
                msg = f"num_functions must be integer, but got {line_str[:10]}"
                raise ValueError(msg) from e
            return num_functions

        def is_dirac_version_less_than_21():
            if self.is_less_than_dirac_21:
                return True
            for atom_info in self.functions_info[component_func][symmetry_label][atom_label].values():
                if atom_info.func_idx_dirac21.first <= num_functions <= atom_info.func_idx_dirac21.last:
                    # If the num_functions is include in the range of the current atom_info indices, it means that the current atom is DIRAC 21 or later.
                    return False
            return True

        def need_to_update_current_atom_info():
            first = self.current_atom_info.func_idx_dirac19.first if self.is_less_than_dirac_21 else self.current_atom_info.func_idx_dirac21.first
            last = self.current_atom_info.func_idx_dirac19.last if self.is_less_than_dirac_21 else self.current_atom_info.func_idx_dirac21.last
            if first <= num_functions <= last:
                return False
            return True

        num_functions = get_func_no()
        component_func = "large" if line_str[10] == "L" else ("small" if line_str[10] == "S" else "")  # CLS
        symmetry_label = line_str[12:15].strip()  # REP (e.g. "Ag "), symmetry_label="Ag"
        atom_label = line_str[15:18].strip()  # NAMN (e.g. "Cm "), atom_labe="Cm"
        gto_type = line_str[18:22].strip()  # GTOTYP (e.g. "s   "), gto_type="s"
        label = symmetry_label + atom_label

        if need_to_update_current_atom_info():
            self.is_less_than_dirac_21 = is_dirac_version_less_than_21()
            found_next_atom_info = False
            for atom_info in self.functions_info[component_func][symmetry_label][atom_label].values():
                start = atom_info.func_idx_dirac19.first if self.is_less_than_dirac_21 else atom_info.func_idx_dirac21.first
                end = atom_info.func_idx_dirac19.last if self.is_less_than_dirac_21 else atom_info.func_idx_dirac21.last
                if start <= num_functions <= end:
                    found_next_atom_info = True
                    self.current_atom_info = fast_deepcopy_pickle(atom_info)
                    self.used_atom_info[label] = atom_info
                    break
            if not found_next_atom_info:
                msg = f"The corresponding atom_info is not found in functions_info[{component_func}][{symmetry_label}][{atom_label}],\
                    list[atom_info] = {list(self.functions_info[component_func][symmetry_label][atom_label].values())}"
                raise Exception(msg)

        self.current_atom_info.decrement_function(gto_type)
        self.data_mo.add_coefficient(get_coefficient(line_str, self.functions_info, self.current_atom_info.idx_within_same_atom))

    def add_current_mo_data_to_data_all_mo(self) -> None:
        self.data_mo.fileter_coefficients_by_threshold()
        # add current MO data to data_all_mo
        # create a new DataMO object using pickle
        copy_data_mo = fast_deepcopy_pickle(self.data_mo)
        if self.is_electronic:
            self.data_all_mo.electronic.append(copy_data_mo)
            cur_sym = self.mo_sym_type
            if args.for_generator:
                self.eigenvalues.energies_used[cur_sym][self.data_mo.eigenvalue_no] = True
        else:
            self.data_all_mo.positronic.append(copy_data_mo)
        debug_print(f"End of reading {self.data_mo.eigenvalue_no}th MO")

    def read_privec_data_wrapper(self):
        """Read coefficients from the output file of DIRAC and store them in data_all_mo.
        This function is intended to wrap the processing of
        read_privec_data between the single-process and multiprocess versions.
        """

        def merge_energies_used(eigenvalues: Eigenvalues) -> None:
            """If multi-process version is used, eigenvalues.energies_used is stored in different memory space.
            Therefore, we need to merge eigenvalues.energies_used of each process.

            Args:
                eigenvalues (Eigenvalues): eigenvalues.energies_used of each process is stored in this variable.
            """
            for sym_type_key, val in eigenvalues.energies_used.items():
                for eigenvalue_no, is_found in val.items():
                    if is_found:
                        self.eigenvalues.energies_used[sym_type_key][eigenvalue_no] = True

        num_processes: int = args.parallel
        if num_processes > 1:
            # Multi-process version
            with concurrent.futures.ProcessPoolExecutor() as executor:
                futures = [executor.submit(self.read_privec_data, i) for i in range(num_processes)]
                result_list = [future.result() for future in concurrent.futures.as_completed(futures)]
            result: Tuple[DataAllMO, Eigenvalues]
            for result in result_list:
                data_all_mo = result[0]
                self.data_all_mo.electronic.extend(data_all_mo.electronic)
                self.data_all_mo.positronic.extend(data_all_mo.positronic)
                eigenvalues = result[1]
                merge_energies_used(eigenvalues)
        else:
            # Single-process version
            self.read_privec_data()

        if args.for_generator:
            self.fill_non_moltra_range_electronic_eigenvalues()
        self.data_all_mo.sort_mo_sym_type()

    def fill_non_moltra_range_electronic_eigenvalues(self):
        self.is_electronic = True
        for sym_type_key, val in self.eigenvalues.energies_used.items():
            self.mo_sym_type = sym_type_key
            for eigenvalue_no, used in val.items():
                if not used:
                    self.data_mo.reset()
                    self.data_mo.eigenvalue_no = eigenvalue_no
                    self.data_mo.mo_info = self.get_mo_info(eigenvalue_no)
                    self.data_mo.sym_type = sym_type_key
                    self.data_mo.mo_energy = self.eigenvalues.energies[sym_type_key][eigenvalue_no]
                    copy_data_mo = fast_deepcopy_pickle(self.data_mo)
                    self.data_all_mo.electronic.append(copy_data_mo)



import os
import re
import subprocess
from pathlib import Path
from typing import List

import pytest


class Env:
    def __init__(self, input_filename: str, options: str, ref_filename: str = "", result_filename: str = "") -> None:
        self.test_path = Path(__file__).resolve().parent
        self.ref_filepath = Path.joinpath(self.test_path, "references", ref_filename)
        self.result_filepath = Path.joinpath(self.test_path, "results", result_filename)
        self.input_filepath = Path.joinpath(self.test_path, "data", input_filename)
        if ref_filename == "":
            self.command: str = f"sum_dirac_dfcoef -i {self.input_filepath} {options}"
        else:
            self.command: str = f"sum_dirac_dfcoef -i {self.input_filepath} -o {self.result_filepath} {options}"


def get_output_list(filepath: Path) -> List[List[str]]:
    return [re.split(" +", line.rstrip("\r\n")) for line in list(open(filepath).read().splitlines())]


@pytest.mark.parametrize(
    "ref_filename, result_filename, input_filename, options",
    # fmt: off
    [
        ("ref.Ar.compress.out"                      , "result.Ar.compress.out"                      , "Ar_Ar.out"                    , "-d 15 -c"),
        ("ref.Ar.no_sort.compress.out"              , "result.Ar.no_sort.compress.out"              , "Ar_Ar.out"                    , "-d 15 --no-sort -c"),
        ("ref.N2.compress.out"                      , "result.N2.compress.out"                      , "N2_N2.out"                    , "-d 15 -c"),
        ("ref.N2.compress.positronic.out"           , "result.N2.compress.positronic.out"           , "N2_N2.out"                    , "-d 15 -pc"),
        ("ref.N2.compress.all.out"                  , "result.N2.compress.all.out"                  , "N2_N2.out"                    , "-d 15 -ac"),
        ("ref.N2.no_sort.compress.out"              , "result.N2.no_sort.compress.out"              , "N2_N2.out"                    , "-d 15 --no-sort -c"),
        ("ref.N2.no_sort.compress.positronic.out"   , "result.N2.no_sort.compress.positronic.out"   , "N2_N2.out"                    , "-d 15 --no-sort -pc"),
        ("ref.N2.no_sort.compress.all.out"          , "result.N2.no_sort.compress.all.out"          , "N2_N2.out"                    , "-d 15 --no-sort -ac"),
        ("ref.uo2.no_sort.compress.out"             , "result.uo2.no_sort.compress.out"             , "x2c_uo2_238.out"              , "-d 15 --no-sort -c"),
        ("ref.ucl4.compress.out"                    , "result.ucl4.compress.out"                    , "x2c_ucl4.out"                 , "-d 15 -g"),
        ("ref.ucl4.no_sort.compress.out"            , "result.ucl4.no_sort.compress.out"            , "x2c_ucl4.out"                 , "-d 15 --no-sort -c"),
        ("ref.Cm3+_phen.compress.out"               , "result.Cm3+_phen.compress.out"               , "x2c_Cm3+_phen.out"            , "-d 15 -g"),
        ("ref.Cm3+_phen_reorder.compress.out"       , "result.Cm3+_phen_reorder.compress.out"       , "x2c_Cm3+_phen_reorder.out"    , "-d 15 -g"),
        ("ref.H2.no_scf.compress.out"               , "result.H2.no_scf.compress.out"               , "H2.noscf_H2.out"              , "-d 15 -c --no-scf"),
        ("ref.H2O.invalid.eigpri.compress.out"      , "result.H2O.invalid.eigpri.compress.out"      , "H2O.invalid.eigpri.out"       , "-d 15 -c --no-scf"),
        ("ref.methane.whitespace.compress.out"      , "result.methane.whitespace.compress.out"      , "methane.whitespace_mol.out"   , "-d 15 -c"),
        # multiprocess (should be the same as the single process case)
        ("ref.ucl4.compress.out"                    , "result.ucl4.compress.multi-process.out"      , "x2c_ucl4.out"                 , "-j2 -d 15 -g"),
        # DIRAC 19 UO2 x2c (https://github.com/RQC-HU/sum_dirac_dfcoef/issues/93#issue-2164290127)
        ("ref.uo2.compress.out"                     , "result.uo2.compress.out"                     , "x2c_uo2_238.out"              , "-d 15 -g"),
        # DIRAC 19 N2 4component (https://github.com/RQC-HU/sum_dirac_dfcoef/issues/93#issuecomment-1976947902)
        ("ref.N2.dirac19.compress.out"              , "result.N2.dirac19.compress.out"              , "dirac19_N2_N2.out"            , "-d 15 -g"),
        # F2 (https://github.com/RQC-HU/sum_dirac_dfcoef/issues/92)
        ("ref.F2.compress.out"                      , "result.F2.compress.out"                      , "F2_F2.out"                    , "-d 15 -g"),
        # CO (No inversion symmetry linear molecule, https://github.com/RQC-HU/sum_dirac_dfcoef/issues/95)
        ("ref.CO.compress.out"                      , "result.CO.compress.out"                      , "CO_CO.out"                    , "-d 15 -g"),
        # Explicitly set **MOLTRA > .SCHEME option (https://github.com/RQC-HU/sum_dirac_dfcoef/issues/108)
        ("ref.N2.scheme4.dirac22.compress.out"      , "result.N2.scheme4.dirac22.compress.out"      , "dirac22_scheme4_N2.out"       , "-d 15 -g"),
        ("ref.N2.scheme6.dirac23.compress.out"      , "result.N2.scheme6.dirac23.compress.out"      , "dirac23_scheme6_N2.out"       , "-d 15 -g"),
        # Ignore magnetic quantum number label (https://github.com/RQC-HU/sum_dirac_dfcoef/issues/115)
        ("ref.ucl4.ignore_ml.compress.out"          , "result.ucl4.ignore_ml.compress.out"          , "x2c_ucl4.out"                 , "-d 15 -g --ignore-ml"),
        # Ignore symmetry label (https://github.com/RQC-HU/sum_dirac_dfcoef/issues/117)
        ("ref.uo2.ignore_sym.compress.out"          , "result.uo2.ignore_sym.compress.out"          , "x2c_uo2_238.out"              , "-d 15 -c --ignore-sym"),
        # Both ignore labels
        ("ref.ucl4.ignore_ml_and_sym.compress.out"  , "result.ucl4.ignore_ml_and_sym.compress.out"  , "x2c_ucl4.out"                 , "-d 15 -g --ignore-ml --ignore-sym"),
    ]
    # fmt: on
)
def test_sum_dirac_dfcoeff_compress(ref_filename: str, result_filename: str, input_filename: str, options: str):
    env = Env(input_filename, options, ref_filename, result_filename)
    os.chdir(env.test_path)
    print(f"{env.test_path} test start...\ncommand: {env.command}")
    subprocess.run(env.command.split(), encoding="utf-8", check=True)

    ref_list: List[List[str]] = get_output_list(env.ref_filepath)
    result_list: List[List[str]] = get_output_list(env.result_filepath)
    # File should have the same number of lines
    assert len(ref_list) == len(result_list), f"Number of lines in {ref_filename}(={len(ref_list)}) and {result_filename}(={len(result_list)}) are different."
    threshold: float = 1e-10
    header = True
    for line_idx, (ref, out) in enumerate(zip(ref_list, result_list)):
        if header:
            assert ref == out
            if len(ref) == 1 and len(out) == 1:  # End header
                header = False
        else:
            if len(ref) < 2 or len(out) < 2:
                continue
            # ref[0]: irrep, ref[1]: energy order index in the irrep, ref[2]: energy, ref[3:]: Symmetry value and coefficient
            # (e.g.) E1u 19 -8.8824415703374 B3uUpx 49.999172476298732 B2uUpy 49.999172476298732
            assert ref[0] == out[0], f"irrep in line {line_idx} of {ref_filename} and {result_filename} are different."
            assert ref[1] == out[1], f"Energy order index in line {line_idx} of {ref_filename} and {result_filename} are different."
            assert float(ref[2]) == pytest.approx(float(out[2]), abs=threshold), f"Energy in line {line_idx} of {ref_filename} and {result_filename} are different."
            for idx, (ref_val, out_val) in enumerate(zip(ref[3:], out[3:])):
                if idx % 2 == 0:
                    assert ref_val == out_val, f"Symmetry value in line {line_idx} of {ref_filename} and {result_filename} are different."
                else:
                    assert float(ref_val) == pytest.approx(
                        float(out_val), abs=threshold
                    ), f"Contribution of the AO in the MO in line {line_idx} of {ref_filename} and {result_filename} are different."


@pytest.mark.parametrize(
    "ref_filename, result_filename, input_filename, options",
    # fmt: off
    [
        ("ref.Ar.out"                       , "result.Ar.out"                       , "Ar_Ar.out"                       , "-d 15"),
        ("ref.Ar.no_sort.out"               , "result.Ar.no_sort.out"               , "Ar_Ar.out"                       , "-d 15 --no-sort"),
        ("ref.N2.out"                       , "result.N2.out"                       , "N2_N2.out"                       , "-d 15"),
        ("ref.N2.positronic.out"            , "result.N2.positronic.out"            , "N2_N2.out"                       , "-d 15 -p"),
        ("ref.N2.all.out"                   , "result.N2.all.out"                   , "N2_N2.out"                       , "-d 15 -a"),
        ("ref.N2.no_sort.out"               , "result.N2.no_sort.out"               , "N2_N2.out"                       , "-d 15 --no-sort"),
        ("ref.N2.no_sort.positronic.out"    , "result.N2.no_sort.positronic.out"    , "N2_N2.out"                       , "-d 15 --no-sort -p"),
        ("ref.N2.no_sort.all.out"           , "result.N2.no_sort.all.out"           , "N2_N2.out"                       , "-d 15 --no-sort -a"),
        ("ref.uo2.special.out"              , "result.uo2.special.out"              , "special_exit_condition_UO2.out"  , "-d 15"),
        ("ref.uo2.out"                      , "result.uo2.out"                      , "x2c_uo2_238.out"                 , "-d 15"),
        ("ref.uo2.no_sort.out"              , "result.uo2.no_sort.out"              , "x2c_uo2_238.out"                 , "-d 15 --no-sort"),
        ("ref.ucl4.out"                     , "result.ucl4.out"                     , "x2c_ucl4.out"                    , "-d 15"),
        ("ref.ucl4.no_sort.out"             , "result.ucl4.no_sort.out"             , "x2c_ucl4.out"                    , "-d 15 --no-sort"),
        ("ref.Cm3+_phen.out"                , "result.Cm3+_phen.out"                , "x2c_Cm3+_phen.out"               , "-d 15"),
        # multiprocess (should be the same as the single process case)
        ("ref.uo2.out"                      , "result.uo2.multi-process.out"        , "x2c_uo2_238.out"                 , "-j2 -d 15"),
    ]
    # fmt: on
)
def test_sum_dirac_dfcoeff(ref_filename: str, result_filename: str, input_filename: str, options: str):
    env = Env(input_filename, options, ref_filename, result_filename)
    os.chdir(env.test_path)
    print(f"{env.test_path} test start...\ncommand: {env.command}")
    subprocess.run(env.command.split(), encoding="utf-8", check=True)

    ref_list: List[List[str]] = get_output_list(env.ref_filepath)
    result_list: List[List[str]] = get_output_list(env.result_filepath)
    # File should have the same number of lines
    assert len(ref_list) == len(result_list), f"Number of lines in {ref_filename}(={len(ref_list)}) and {result_filename}(={len(result_list)}) are different."
    threshold: float = 1e-10
    header = True
    for line_idx, (ref, out) in enumerate(zip(ref_list, result_list)):
        if header:
            assert ref == out
            if len(ref) == 1 and len(out) == 1:  # End header
                header = False
        else:
            if len(ref) < 2 or len(out) < 2:
                continue
            if "%" in ref[-1]:
                ref_value = float(ref[-2])
                out_value = float(out[-2])
                ref_list_str = " ".join(ref[:-2])
                out_list_str = " ".join(out[:-2])
            else:
                ref_value = float(ref[-1])
                out_value = float(out[-1])
                ref_list_str = " ".join(ref[:-1])
                out_list_str = " ".join(out[:-1])
            assert ref_list_str == out_list_str, f"line {line_idx}: {ref_list_str} != {out_list_str}\nref: {ref_list[line_idx]}\nout:{result_list[line_idx]}"
            assert ref_value == pytest.approx(out_value, abs=threshold), f"line {line_idx}: {ref_value} != {out_value}\nref: {ref_list[line_idx]}\nout:{result_list[line_idx]}"


@pytest.mark.parametrize(
    "input_filename, options, expected_error_message",
    # fmt: off
    [
        ("H2O.invalid.eigpri.out"  , "-g -d 15"    , "Your .EIGPRI option in your DIRAC input file is invalid!"),
        ("H2.noscf_H2.out"         , "-g -d 15"    , "Cannot find SCF calculation settings"),
        ("Ar_Ar.out"               , "-g --no-scf" , "-g/--for-generator and --no-scf options cannot be set at the same time"),
        ("Ar_Ar.out"               , "-g -p"       , "-g/--for-generator and -p/--positronic-write options cannot be set at the same time"),
    ],
    # fmt: on
)
def test_invalid_option_raise_error(input_filename: str, options: str, expected_error_message: str):
    env = Env(input_filename, options)
    os.chdir(env.test_path)
    print(f"{env.test_path} test start...\ncommand: {env.command}")
    # Run the command and capture the output
    p = subprocess.run(env.command.split(), encoding="utf-8", stderr=subprocess.PIPE, check=False)
    assert p.returncode != 0, "Failed: Command should fail but return code is 0."
    assert expected_error_message in p.stderr, f"Failed: Expected error message not found in output:\nexptected: {expected_error_message}\nstderr: {p.stderr}"


@pytest.mark.parametrize(
    "ref_filename, result_filename, input_filename, options, expected_warning_message",
    # fmt: off
    [
        ("ref.Ar.no_vector_print_data.out", "result.Ar.no_vector_print_data.out", "no_vector_print_data_Ar_Ar.out", "-d 15 -c", "WARNING: The next title is detected before the end of reading coefficients."),  # noqa: E501
    ],
    # fmt: on
)
def test_no_vector_print_data(ref_filename: str, result_filename: str, input_filename: str, options: str, expected_warning_message: str):
    env = Env(input_filename, options, ref_filename, result_filename)
    os.chdir(env.test_path)
    print(f"{env.test_path} test start...\ncommand: {env.command}")
    process = subprocess.run(env.command.split(), encoding="utf-8", check=True, stdout=subprocess.PIPE)

    # Check the warning message
    assert expected_warning_message in process.stdout, f"Failed: Expected warning message not found in output:\nexptected: {expected_warning_message}\nstdout: {process.stdout}"

    # Check the output file
    ref_list: List[List[str]] = get_output_list(env.ref_filepath)
    result_list: List[List[str]] = get_output_list(env.result_filepath)
    # File should have the same number of lines
    assert len(ref_list) == len(result_list), f"Number of lines in {ref_filename}(={len(ref_list)}) and {result_filename}(={len(result_list)}) are different."
    threshold: float = 1e-10
    header = True
    for line_idx, (ref, out) in enumerate(zip(ref_list, result_list)):
        if header:
            assert ref == out
            if len(ref) == 1 and len(out) == 1:  # End header
                header = False
        else:
            if len(ref) < 2 or len(out) < 2:
                continue
            # ref[0]: irrep, ref[1]: energy order index in the irrep, ref[2]: energy, ref[3:]: Symmetry value and coefficient
            # (e.g.) E1u 19 -8.8824415703374 B3uUpx 49.999172476298732 B2uUpy 49.999172476298732
            assert ref[0] == out[0], f"irrep in line {line_idx} of {ref_filename} and {result_filename} are different."
            assert ref[1] == out[1], f"Energy order index in line {line_idx} of {ref_filename} and {result_filename} are different."
            assert float(ref[2]) == pytest.approx(float(out[2]), abs=threshold), f"Energy in line {line_idx} of {ref_filename} and {result_filename} are different."
            for idx, (ref_val, out_val) in enumerate(zip(ref[3:], out[3:])):
                if idx % 2 == 0:
                    assert ref_val == out_val, f"Symmetry value in line {line_idx} of {ref_filename} and {result_filename} are different."
                else:
                    assert float(ref_val) == pytest.approx(
                        float(out_val), abs=threshold
                    ), f"Contribution of the AO in the MO in line {line_idx} of {ref_filename} and {result_filename} are different."


def test_version_option():
    command = "sum_dirac_dfcoef -v"
    p = subprocess.run(command.split(), encoding="utf-8", check=True, stdout=subprocess.PIPE)
    # p.stdout = x.y.z\n
    out_version_str = p.stdout.rstrip("\n")
    about_file = Path(__file__).resolve().parent.parent / "src/sum_dirac_dfcoef/__about__.py"
    # readline() => __version__ = "x.y.z"
    # split()[-1] => "x.y.z"
    # replace() => x.y.z
    ref_version_str = open(about_file).readline().split()[-1].replace('"', "")
    assert out_version_str == ref_version_str


def test_help_option():
    command = "sum_dirac_dfcoef -h"
    p = subprocess.run(command.split(), check=True)
    assert p.returncode == 0



import re
from bisect import bisect_left, bisect_right
from collections import OrderedDict
from io import TextIOWrapper
from typing import List
from typing import OrderedDict as ODict

from sum_dirac_dfcoef.data import DataAllMO, DataMO
from sum_dirac_dfcoef.eigenvalues import Eigenvalues
from sum_dirac_dfcoef.electron_num import get_electron_num_from_input, get_electron_num_from_scf_field
from sum_dirac_dfcoef.moltra import MoltraInfo
from sum_dirac_dfcoef.scheme import Scheme


class HeaderInfo:
    """Class to store header information for the sum_dirac_dfcoef module.

    Attributes:
        moltra_info (MoltraInfo): Moltra information
        eigenvalues (Eigenvalues): Eigenvalues
        point_group (str): Point group of the molecule (e.g.) "C2v"
        electrons (int): Number of electrons
    """

    def __init__(self):
        self.moltra_info = MoltraInfo()
        self.eigenvalues = Eigenvalues()
        self.point_group = ""
        self.scheme = Scheme()
        self.electrons = 0

    def read_header_info(self, dirac_output: TextIOWrapper) -> None:
        """Read the header information from the output file of DIRAC

        Args:
            dirac_output (TextIOWrapper): Output file of DIRAC

        Returns:
            None: class attributes are updated
        """
        dirac_output.seek(0)
        self.__read_electron_number(dirac_output)
        dirac_output.seek(0)
        self.__validate_eigpri_option(dirac_output)
        dirac_output.seek(0)
        self.__read_scheme(dirac_output)
        dirac_output.seek(0)
        self.__read_moltra(dirac_output)
        self.__read_point_group(dirac_output)
        self.__read_eigenvalues(dirac_output)
        self.__duplicate_moltra_str()

    def __read_point_group(self, dirac_output: TextIOWrapper) -> None:
        symgrp_section = False
        err_msg = "The symmetry group is not found in the output file of DIRAC."
        for line in dirac_output:
            if "SYMGRP" in line:
                symgrp_section = True
                continue
            if symgrp_section:
                if "Represented as" in line or "Point group:" in line:
                    self.point_group = line.split()[-1]
                    break
                else:
                    line_without_space = line.strip().strip("\r\n")
                    if all("*" == char for char in line_without_space) and len(line_without_space) > 0:
                        raise ValueError(err_msg)
        if self.point_group == "":
            raise ValueError(err_msg)

    def __read_electron_number(self, dirac_output: TextIOWrapper) -> None:
        self.electrons = get_electron_num_from_input(dirac_output)
        if self.electrons == 0:
            self.electrons = get_electron_num_from_scf_field(dirac_output)

    def __validate_eigpri_option(self, dirac_output: TextIOWrapper) -> None:
        self.eigenvalues.validate_eigpri_option(dirac_output)

    def __read_scheme(self, dirac_output: TextIOWrapper) -> None:
        self.scheme.get_scheme_num_from_input(dirac_output)

    def __read_eigenvalues(self, dirac_output: TextIOWrapper) -> None:
        self.eigenvalues.get_eigenvalues(dirac_output)

    def __read_moltra(self, dirac_output: TextIOWrapper) -> None:
        self.moltra_info.read_moltra_section(dirac_output)

    def __duplicate_moltra_str(self) -> None:
        # Duplicate the moltra range string if it is not enough
        if self.moltra_info.is_default:
            # Set the default range string
            for _ in range(len(self.eigenvalues.shell_num)):
                self.moltra_info.range_str.append("ENERGY -20.0 10.0 1.0")
        if len(self.moltra_info.range_str) != len(self.eigenvalues.shell_num):
            # one-line input is allowed, duplicate the first line
            # https://gitlab.com/dirac/dirac/-/blob/ea717cdb294035d8af3ebe2b1e00cf94f1c1a6b7/src/moltra/trainp.F#L592-600
            for _ in range(len(self.eigenvalues.shell_num) - len(self.moltra_info.range_str)):
                self.moltra_info.range_str.append(self.moltra_info.range_str[0])

    def calculate_moltra_idx_range(self, data_all_mo: DataAllMO) -> None:
        keys = list(self.eigenvalues.shell_num.keys())
        for i, item in enumerate(self.moltra_info.range_str):
            symmetry_type = keys[i]
            if "ALL" == item.upper():
                self.moltra_info.range_dict[symmetry_type] = f"1..{len(self.eigenvalues.energies[symmetry_type])}"
            elif "ENERGY" in item.upper():
                self.moltra_info.range_dict[symmetry_type] = self.__parse_energy_str(item, symmetry_type, data_all_mo)
            else:
                self.moltra_info.range_dict[symmetry_type] = self.__parse_range_str(item, symmetry_type)

    def __parse_energy_str(self, energy_str: str, symmetry_type: str, data_all_mo: DataAllMO) -> str:
        """Parse the energy string

        Args:
            energy_str (str): Energy string

        Returns:
            str: Range string
        """

        def get_min_energy_idx(energies: List[float], min_energy: float, step: float) -> int:
            # Find the index of the minimum energy without exceeding the step
            cur_idx = bisect_left(energies, min_energy)
            cur_min_energy = energies[cur_idx]
            # Search for the minimum energy index
            while cur_idx > 0:
                next_energy = energies[cur_idx - 1]
                if abs(cur_min_energy - next_energy) > step:
                    break  # Found the minimum energy index
                cur_min_energy = next_energy
                cur_idx -= 1
            return cur_idx

        def get_max_energy_idx(energies: List[float], max_energy: float, step: float) -> int:
            # Find the index of the maximum energy without exceeding the step
            cur_idx = bisect_right(energies, max_energy)
            cur_max_energy = energies[cur_idx - 1]
            while cur_idx < len(energies):
                next_energy = energies[cur_idx]
                if abs(next_energy - cur_max_energy) > step:
                    break  # Found the maximum energy index
                cur_max_energy = next_energy
                cur_idx += 1
            return cur_idx

        def find_min_sym_idx() -> int:
            """Find the minimum index of the symmetry type

            Returns:
                int: Minimum index of the symmetry type
            """

            def bisect_l(data: List[DataMO], key: str) -> int:
                min_idx, max_idx = 0, len(data)
                while min_idx < max_idx:
                    mid_idx = (min_idx + max_idx) // 2
                    if data[mid_idx].sym_type < key:
                        min_idx = mid_idx + 1
                    else:
                        max_idx = mid_idx
                return min_idx

            key = symmetry_type
            min_idx = bisect_l(data_all_mo.electronic, key)
            return min_idx

        def create_within_moltra_dict(start_idx: int, end_idx: int) -> OrderedDict:
            """Create the within_moltra dictionary

            Returns:
                OrderedDict: within_moltra dictionary
            """
            within_moltra = OrderedDict(sorted({mo.eigenvalue_no: False for mo in data_all_mo.electronic}.items()))
            for idx in range(start_idx, end_idx):
                within_moltra[data_all_mo.electronic[idx].eigenvalue_no] = True
            return within_moltra

        def create_energy_str(within_moltra: ODict[int, bool]) -> str:
            """Create the energy string from the input boolean dictionary

            Args:
                within_moltra (ODict[int, bool]): Stores the boolean value of whether the eigenvalue is used or not. True: used, False: not used.

            Returns:
                energy_str (str): Energy string can be used in the DIRAC input file. (e.g.) "10..180,200..300,400..500"
            """
            energy_str = ""
            cur_mo_num = 0
            start_mo_num = cur_mo_num
            left = True
            ser_end = False
            for eigenvalue_no, is_used in within_moltra.items():
                cur_mo_num = eigenvalue_no
                if is_used:
                    if left:  # Add the first number of the series
                        if energy_str == "":
                            energy_str += f"{cur_mo_num}"
                        else:  # Add the comma if the string is not empty to separate the series
                            energy_str += f",{cur_mo_num}"
                        start_mo_num = cur_mo_num
                        left = False
                        ser_end = True
                elif ser_end:  # Add the last number of the series
                    if cur_mo_num > start_mo_num + 1:
                        energy_str += f"..{cur_mo_num - 1}"
                    left = True
                    ser_end = False
            if ser_end:  # Add the last number of the series if the series is not ended after the loop
                energy_str += f"..{cur_mo_num}"
            return energy_str

        start_sym_idx = find_min_sym_idx()
        energy_str = energy_str.upper().replace("ENERGY", "")
        min_energy, max_energy, step = map(float, energy_str.split())
        if min_energy > max_energy:
            msg = f"The minimum energy is larger than the maximum energy: {min_energy} > {max_energy}"
            raise ValueError(msg)
        energies: List[float] = sorted(self.eigenvalues.energies[symmetry_type].values())
        min_energy_idx = get_min_energy_idx(energies, min_energy, step)
        max_energy_idx = get_max_energy_idx(energies, max_energy, step)

        within_moltra = create_within_moltra_dict(min_energy_idx + start_sym_idx, max_energy_idx + start_sym_idx)
        energy_str = create_energy_str(within_moltra)
        return energy_str

    def __parse_range_str(self, range_str: str, symmetry_type: str) -> str:
        """Parse the range string
        (e.g.) "10..180, 200..300, 400..oo" => "10..180,200..300,400..500"

        Args:
            range_str (str): Range string

        Returns:
            str: Range string
        """
        # [-]?(?:[0-9]+|oo) => -oo or oo or integer
        # \.{2} => ..
        regex = r"[-]?(?:[0-9]+|oo)\.{2}[-]?(?:[0-9]+|oo)"
        items: List[str] = re.findall(regex, range_str)
        ret_list: List[str] = []
        for item in items:
            item_li = item.replace("..", " ").split()
            range_li = [1 if item == "-oo" else len(self.eigenvalues.energies[symmetry_type]) if item == "oo" else int(item) for item in item_li]
            if range_li[0] > range_li[1]:
                msg = f"The minimum index is larger than the maximum index: {range_li[0]} > {range_li[1]}\n\
your input: {range_str}, invalid input part: {item}"
                raise ValueError(msg)
            ret_list.append(f"{range_li[0]}..{range_li[1]}")
        return ",".join(ret_list)



# If user uses this package as a script, the following code is executed. Otherwise(used as a module), __init__.py is executed.
import sys

if __name__ == "__main__":
    from sum_dirac_dfcoef.sum_dirac_dfcoef import main

    sys.exit(main())



import re
from io import TextIOWrapper

from sum_dirac_dfcoef.utils import (
    delete_dirac_input_comment_out,
    is_dirac_input_line_should_be_skipped,
    is_dirac_input_section,
    is_end_dirac_input_field,
    is_start_dirac_input_field,
    space_separated_parsing,
    space_separated_parsing_upper,
)


def get_electron_num_from_input(dirac_output: TextIOWrapper) -> int:
    """If users calculate SCF with open shell they must explicitly write the OPEN SHELL and CLOSED SHELL keywords
    in the input file. Therefore, we can get the electron number from the input file.

        Args:
            dirac_output (TextIOWrapper): Output file of DIRAC

        Returns:
            int: The number of electrons in the system
    """

    def get_a_natural_number(word: str) -> int:
        # OPEN SHELL and CLOSED SHELL electron number settings
        # must be written as a natural number (negative number is not allowed)
        regex_number = r"[-]?[0-9]+"
        number = int(re.search(regex_number, word).group())
        if number < 0:
            msg = "The number of electrons in OPEN SHELL and CLOSED SHELL must be a natural number.\n\
But we found a negative number in your DIRAC input file.\n\
Please check your DIRAC input file and try again.\n"
            raise ValueError(msg)
        return number

    electron_num: int = 0
    is_closed_shell_section: bool = False
    is_openshell_section: bool = False
    num_of_open_shell: int = 0
    is_reach_input_field: bool = False
    is_scf_found: bool = False
    scf_detail_section: bool = False
    for line in dirac_output:
        no_comment_out_line = delete_dirac_input_comment_out(line)
        words = space_separated_parsing_upper(no_comment_out_line)

        if is_dirac_input_line_should_be_skipped(words):
            continue

        if is_start_dirac_input_field(no_comment_out_line):
            is_reach_input_field = True
            continue

        if is_end_dirac_input_field(no_comment_out_line):
            break  # end of input field

        if is_reach_input_field:
            if ".SCF" in words[0]:
                is_scf_found = True

            if is_dirac_input_section(words[0]):
                if "*SCF" in words[0]:
                    scf_detail_section = True
                else:
                    scf_detail_section = False

            if scf_detail_section:
                if is_openshell_section:
                    # open shell format
                    # https://diracprogram.org/doc/master/manual/wave_function/scf.html#open-shell
                    # .OPEN SHELL
                    # num_of_open_shell
                    # num_of_elec/irrep1_num_spinor irrep2_num_spinor ...
                    # We want to get only num_of_elec
                    if num_of_open_shell == 0:
                        num_of_open_shell = get_a_natural_number(words[0])
                    else:
                        electron_num += get_a_natural_number(words[0])
                        num_of_open_shell -= 1  # Got an electron_num, so decrease the num_of_open_shell
                        if num_of_open_shell == 0:  # Got all open shell electron_num
                            is_openshell_section = False

                if is_closed_shell_section:
                    # closed shell format
                    # https://diracprogram.org/doc/master/manual/wave_function/scf.html#closed-shell
                    # .CLOSED SHELL
                    # irrep1_num_spinor irrep2_num_spinor ...
                    for word in words:
                        electron_num += get_a_natural_number(word)
                    is_closed_shell_section = False

                # .CLOSED SHELL
                if ".CLOSED" == words[0] and "SHELL" in words[1]:
                    is_closed_shell_section = True

                # .OPEN SHELL
                if ".OPEN" == words[0] and "SHELL" in words[1]:
                    is_openshell_section = True
    if not is_scf_found:
        msg = "\nCannot find SCF calculation settings from your DIRAC input file written in your output file.\n\
we cannot get information about the electron number and orbital energy without SCF calculation.\n\
So we cannot continue this program because we need electron number and orbital energies to summarize DIRAC output.\n\
Please check your DIRAC input file and try again.\n\
If you want to use this program by using the output file without SCF calculation, please use --no-scf option.\n\
But you cannot use the output using --no-scf option to dcaspt2_input_generator program.\n"
        raise ValueError(msg)
    return electron_num


def get_electron_num_from_scf_field(dirac_output: TextIOWrapper) -> int:
    # https://gitlab.com/dirac/dirac/-/blob/79e6b9e27cf8018999ddca2aa72247ccfb9d2a2d/src/dirac/dirrdn.F#L2127
    # find "i.e. no. of electrons ="
    is_wave_function_module_reached: bool = False
    for line in dirac_output:
        words = space_separated_parsing(line)
        if "Wave function module" in line:
            is_wave_function_module_reached = True
            continue

        if is_wave_function_module_reached:
            if "i.e. no. of electrons" in line:
                # ["i.e.", "no.", "of", "electrons", "=", number]
                return int(words[5])
    msg = "\nCannot find electron number from your DIRAC output file.\n\
we cannot get information about the electron number and orbital energy without SCF calculation.\n\
So we cannot continue this program because we need electron number and orbital energies to summarize DIRAC output.\n\
Please check your DIRAC input file and try again.\n\
If you want to use this program by using the output file without SCF calculation, please use --no-scf option.\n\
But you cannot use the output using --no-scf option to dcaspt2_input_generator program.\n"
    raise ValueError(msg)



import re
from collections import OrderedDict
from enum import Enum, auto
from io import TextIOWrapper
from typing import Dict, List, Optional, Tuple
from typing import OrderedDict as ODict

from sum_dirac_dfcoef.utils import (
    debug_print,
    delete_dirac_input_comment_out,
    is_dirac_input_keyword,
    is_dirac_input_line_should_be_skipped,
    is_dirac_input_section_one_star,
    is_end_dirac_input_field,
    is_start_dirac_input_field,
    space_separated_parsing,
    space_separated_parsing_upper,
)


class StageEigenvalues(Enum):
    INIT = auto()
    SEARCH_EIGENVALUES_HEADER = auto()
    SEARCH_PRINT_TYPE = auto()
    EIGENVALUES_READ = auto()
    OCCUPATION_INFO_READ = auto()
    WAIT_END = auto()


class StageEIGPRI(Enum):
    INIT = auto()
    SEARCH_SCF_SECTION = auto()
    SEARCH_SCF_DETAIL_SECTION = auto()
    SEARCH_EIGPRI = auto()
    NEXT_LINE_EIGPRI = auto()
    WAIT_END = auto()


# type definition eigenvalues.shell_num
# type shell_num = {
#     "E1g": {
#         "closed": int
#         "open": int
#         "virtual": int
#     },
#     "E1u": {
#         "closed": int
#         "open": int
#         "virtual": int
#     },
# }
class Eigenvalues:
    """This class stores the eigenvalues information after SCF calculation in the DIRAC output file.

    Raises:
        ValueError: Raises error when the .EIGPRI option in the DIRAC input file is invalid. (.EIGPRI 0 1 or .EIGPRI 0 0)
        ValueError: Raises error when occ_idx[item] is not found in omega_list.

    Attributes:
        shell_num (ODict[str, Dict[str, int]]): The number of closed, open, and virtual orbitals for each symmetry type.
        energies (ODict[str, Dict[int, float]]): The eigenvalues for each symmetry type.
        energies_used (ODict[str, Dict[int, bool]]): The flag to check whether a specified index eigenvalue exists in the Vector print data or not.
                                                     Set to True if the eigenvalue is found in the Vector print data.
                                                     (read Vector print data at PrivecProcessor.read_privec_data_wrapper() method)
                                                     This flag is used when filling the eigenvalues info which is not found in the Vector print data.

    """

    shell_num: ODict[str, Dict[str, int]]
    energies: ODict[str, Dict[int, float]]
    energies_used: ODict[str, Dict[int, bool]]

    def __init__(
        self,
        shell_num: Optional[ODict[str, Dict[str, int]]] = None,
        energies: Optional[ODict[str, Dict[int, float]]] = None,
        energies_used: Optional[ODict[str, Dict[int, bool]]] = None,
    ) -> None:
        self.shell_num = shell_num if shell_num is not None else OrderedDict()
        self.energies = energies if energies is not None else OrderedDict()
        self.energies_used = energies_used if energies_used is not None else OrderedDict()

    def __repr__(self) -> str:
        return f"Eigenvalues(shell_num: {self.shell_num}\nenergies: {self.energies}\nenergies_used: {self.energies_used})"

    def setdefault(self, key: str):
        self.shell_num.setdefault(key, {"closed": 0, "open": 0, "virtual": 0, "negative": 0, "positronic": 0})
        self.energies.setdefault(key, {})
        self.energies_used.setdefault(key, {})

    def get_electronic_spinor_num(self, symmetry_type: str) -> int:
        return self.shell_num[symmetry_type]["closed"] + self.shell_num[symmetry_type]["open"] + self.shell_num[symmetry_type]["virtual"]

    def get_eigenvalues(self, dirac_output: TextIOWrapper):
        def is_end_of_read(line) -> bool:
            return True if "HOMO - LUMO" in line else False

        def is_eigenvalue_type_written(words: List[str]) -> bool:
            # closed shell: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1043
            # open shell: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1053
            # virtual eigenvalues: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1064
            # negative energy eigenvalues (only atom or linear molecule case): https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1156
            # positronic eigenvalues (not atom and linear molecule case): https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1073
            if "*" == words[0] and "Closed" == words[1] and "shell," == words[2]:
                return True
            elif "*" == words[0] and "Open" == words[1] and "shell" == words[2]:
                return True
            elif "*" == words[0] and "Virtual" == words[1] and "eigenvalues," == words[2]:
                return True
            elif "*" == words[0] and "Negative" == words[1] and "energy" == words[2] and "eigenvalues," == words[3]:
                return True
            elif "*" == words[0] and "Positronic" == words[1] and "eigenvalues," == words[2]:
                return True
            return False

        def is_occupation_info_header(line: str) -> bool:
            return True if "Occupation in fermion symmetry" in line else False

        def prepare_occupation_info(words: List[str]) -> Tuple[str, Dict[str, int]]:
            current_symmetry_type = words[len(words) - 1]
            occ_idx = {k: 1 for k in omega[current_symmetry_type].keys()}
            return current_symmetry_type, occ_idx

        def get_current_eigenvalue_type(words: List[str]) -> str:
            # words[0] = '*', words[1] = "Closed" or "Open" or "Virtual" or "Negative" or "Positronic"
            current_eigenvalue_type = words[1].lower()
            return current_eigenvalue_type

        def get_symmetry_type_standard(words: List[str]) -> str:
            current_symmetry_type = words[3]
            return current_symmetry_type

        def get_symmetry_type_supersym(line: str) -> str:
            # https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1097-1105
            # FORMAT '(/A,I4,4A,I2,...)'
            # DATA "* Block",ISUB,' in ',FREP(IFSYM),":  ",...
            # ISUB might be **** if ISUB > 9999 or ISUB < -999 because of the format
            # Therefore, find 'in' and ':' in line and get FREP(IFSYM) from the line
            # FREP(IFSYM) is a symmetry type
            in_idx = line.index("in")
            colon_idx = line.index(":")
            current_symmetry_type = line[in_idx + 2 : colon_idx].strip()
            return current_symmetry_type

        def get_omega_str(words: List[str]) -> str:
            if "Omega" in line:
                # * Block   3 in E1u:  Omega =  5/2
                # => 5/2
                omega_str = words[len(words) - 1].replace("=", "").strip()
                return omega_str
            else:
                # * Block   3 in E1u:  p 3/2; -3/2
                # => p 3/2 -3/2
                colon_idx = line.index(":")
                omega_str = line[colon_idx + 1 : len(line) - 1].strip()
                if ";" in omega_str:
                    jval = omega_str.split(";")[0].strip()
                    mjval = omega_str.split(";")[1].strip()
                    omega_str = f"{jval} {mjval}"
                return omega_str

        def supersym_append_energies() -> None:
            for item in omega_list:
                if item not in occ_idx.keys():
                    msg = f"Cannot find {item} in occ_idx.keys()!, occ_idx.keys(): {occ_idx.keys()}"
                    raise ValueError(msg)
                val = omega[current_symmetry_type][item][occ_idx[item]]
                idx = len(self.energies[current_symmetry_type]) + 1
                self.energies[current_symmetry_type][idx] = val
                occ_idx[item] += 1

        def create_splitted_by_slash2_list(line: str) -> List[str]:
            # split by /2 and remove empty string
            # (e.g. 1, atomic  ) "s 1/2 d 3/2 s 1/2" => ["s 1/2", "d 3/2", "s 1/2"]
            # (e.g. 2, molecule) "1/2 1/2 1/2 3/2 1/2 5/2" => ["1/2", "1/2", "1/2", "3/2", "1/2", "5/2"]
            split_by_slash2 = list(filter(None, [item.strip("\r\n") for item in line.split("/2")]))
            split_by_slash2 = [f"{item.strip()}/2" for item in split_by_slash2]
            return split_by_slash2

        def add_mj_info_to_omega_list(line: str, omega_list: List[str]) -> List[str]:
            mj_list = create_splitted_by_slash2_list(line.replace("Mj", "").strip("\r\n"))
            if len(mj_list) != len(omega_list):
                msg = f"len(mj_list) != len(omega_list)\nmj_list: {mj_list}\nomega_list: {omega_list}\nline: {line}\n"
                raise ValueError(msg)
            omega_list = [f"{omega_list[i]} {mj_list[i]}" for i in range(len(mj_list))]
            return omega_list

        def read_eigenvalues(line: str) -> None:
            start_idx = 0
            while True:
                # e.g. -775.202926514  ( 2) => -775.202926514
                regex = r"[-]?[0-9]+\.?[0-9]+"
                match = re.search(regex, line[start_idx:])
                if match is None:
                    break
                val = float(match.group())

                # e.g. -775.202926514  ( 2) => 2
                regex = r"\([ ]*[0-9]+\)"
                match = re.search(regex, line[start_idx:])
                if match is None:
                    break
                # match.group() == ( 2) => [1 : len(match.group()) - 1] == 2
                num = int(match.group()[1 : len(match.group()) - 1])
                self.shell_num[current_symmetry_type][current_eigenvalue_type] += num
                if print_type == "standard":
                    for _ in range(0, num, 2):
                        idx = len(self.energies[current_symmetry_type]) + 1
                        self.energies[current_symmetry_type][idx] = val
                elif print_type == "supersymmetry":
                    for _ in range(0, num, 2):
                        idx = len(omega[current_symmetry_type][omega_str]) + 1
                        omega[current_symmetry_type][omega_str][idx] = val
                start_idx += match.end()

        stage = StageEigenvalues.INIT
        atomic = False
        print_type = ""  # "standard" or "supersymmetry"
        occ_idx: Dict[str, int] = {}
        omega: Dict[str, Dict[str, Dict[int, float]]] = {}
        omega_str = ""  # 1/2 or 3/2 or 5/2 or p 3/2 -3/2 ...
        omega_list = []
        current_eigenvalue_type = ""  # "closed" or "open" or "virtual"
        current_symmetry_type = ""  # "E1g" or "E1u" or "E1" ...
        eigenvalue_type_omega_replacement = {"inactive": "closed", "active": "open", "virtual": "virtual"}

        for line in dirac_output:
            words: List[str] = space_separated_parsing(line)

            if len(words) == 0:
                continue
            elif stage == StageEigenvalues.INIT:
                if "SCF - CYCLE" in line:
                    stage = StageEigenvalues.SEARCH_EIGENVALUES_HEADER
            elif stage == StageEigenvalues.SEARCH_EIGENVALUES_HEADER:
                if "Eigenvalues" == words[0]:
                    stage = StageEigenvalues.SEARCH_PRINT_TYPE
            elif stage == StageEigenvalues.SEARCH_PRINT_TYPE:
                if "*" == words[0] and "Fermion" in words[1] and "symmetry" in words[2]:
                    stage = StageEigenvalues.EIGENVALUES_READ
                    print_type = "standard"
                    current_symmetry_type = get_symmetry_type_standard(words)
                    self.setdefault(current_symmetry_type)
                elif "* Block" in line:
                    stage = StageEigenvalues.EIGENVALUES_READ
                    print_type = "supersymmetry"
                    current_symmetry_type = get_symmetry_type_supersym(line)
                    atomic = ";" in line
                    omega_str = get_omega_str(words)
                    self.setdefault(current_symmetry_type)
                    omega.setdefault(current_symmetry_type, {}).setdefault(omega_str, {})
            elif is_end_of_read(line) or stage == StageEigenvalues.WAIT_END:
                break
            elif stage == StageEigenvalues.EIGENVALUES_READ:
                if print_type == "standard" and "*" == words[0] and "Fermion" in words[1] and "symmetry" in words[2]:
                    current_symmetry_type = get_symmetry_type_standard(words)
                    self.setdefault(current_symmetry_type)
                elif print_type == "supersymmetry" and "* Block" in line:
                    current_symmetry_type = get_symmetry_type_supersym(line)
                    omega_str = get_omega_str(words)
                    self.setdefault(current_symmetry_type)
                    omega.setdefault(current_symmetry_type, {}).setdefault(omega_str, {})
                elif is_eigenvalue_type_written(words):
                    current_eigenvalue_type = get_current_eigenvalue_type(words)
                elif is_occupation_info_header(line):
                    stage = StageEigenvalues.OCCUPATION_INFO_READ
                    current_symmetry_type, occ_idx = prepare_occupation_info(words)
                else:
                    read_eigenvalues(line)
            elif stage == StageEigenvalues.OCCUPATION_INFO_READ:
                if "Occupation of" in line:
                    stage = StageEigenvalues.WAIT_END
                elif is_occupation_info_header(line):
                    current_symmetry_type, occ_idx = prepare_occupation_info(words)
                elif "orbitals" in line:
                    # * Inactive orbitals => inactive
                    occ_type = words[1].lower()
                    # inactive => closed
                    current_eigenvalue_type = eigenvalue_type_omega_replacement[occ_type]
                elif "Mj" in line:
                    omega_list = add_mj_info_to_omega_list(line, omega_list)
                    supersym_append_energies()
                elif atomic:
                    omega_list = create_splitted_by_slash2_list(line)
                else:  # molecular
                    omega_list = create_splitted_by_slash2_list(line)
                    supersym_append_energies()

        for key in self.energies.keys():
            num = len(self.energies[key])
            self.energies_used[key] = {i: False for i in range(1, num + 1)}

        debug_print(f"eigenvalues: {self}")

    def validate_eigpri_option(self, dirac_output: TextIOWrapper):
        """Validate the .EIGPRI option in the DIRAC input file,
        if is not set, it is a valid input
        because only the positive energy eigenvalues are printed as default.

        Args:
            dirac_output (TextIOWrapper): _description_
        """

        stage = StageEIGPRI.INIT
        for line in dirac_output:
            no_comment_out_line = delete_dirac_input_comment_out(line)
            words = space_separated_parsing_upper(no_comment_out_line)
            if is_dirac_input_line_should_be_skipped(words):
                continue

            if is_end_dirac_input_field(no_comment_out_line) or stage == StageEIGPRI.WAIT_END:
                break

            if stage == StageEIGPRI.INIT:
                if is_start_dirac_input_field(no_comment_out_line):
                    stage = StageEIGPRI.SEARCH_SCF_SECTION

            if stage == StageEIGPRI.SEARCH_SCF_SECTION:
                if is_dirac_input_keyword(words[0]) and ".SCF" in words[0]:
                    stage = StageEIGPRI.SEARCH_SCF_DETAIL_SECTION

            if stage == StageEIGPRI.SEARCH_SCF_DETAIL_SECTION:
                if is_dirac_input_section_one_star(words[0]) and "*SCF" in words[0]:
                    stage = StageEIGPRI.SEARCH_EIGPRI

            if stage == StageEIGPRI.SEARCH_EIGPRI:
                if is_dirac_input_keyword(words[0]) and ".EIGPRI" in words[0]:
                    stage = StageEIGPRI.NEXT_LINE_EIGPRI

            if stage == StageEIGPRI.NEXT_LINE_EIGPRI:
                # https://diracprogram.org/doc/master/manual/wave_function/scf.html#eigpri
                if len(words) == 2 and words[0].isdigit() and words[1].isdigit():
                    if int(words[0]) == 0:  # positive energy eigenvalues are not printed
                        msg = f"\nYour .EIGPRI option in your DIRAC input file is invalid!\n\
.EIGPRI\n\
{line}\n\
We cannot get the eigenvalues with your .EIGPRI option.\n\
If you want to use this output file with this program, you must use --no-scf option to skip reading eigenvalues information.\n\
But you cannot use the output using --no-scf option to dcaspt2_input_generator program.\n\
If you want to get the eigenvalues information, please refer the .EIGPRI option in the manual of DIRAC.\n\
https://diracprogram.org/doc/master/manual/wave_function/scf.html#eigpri\n\
You must enable to print out the positive eigenvalues energy.\n"
                        raise ValueError(msg)
                    stage = StageEIGPRI.WAIT_END



import os
import subprocess
from pathlib import Path


def pytest_sessionstart():
    cur_dir = Path.cwd()
    # Change the current directory to the root directory of the package
    test_dir = Path(__file__).resolve().parent
    results_dir = Path.joinpath(test_dir, "results")
    if results_dir.exists():
        # Remove *.out files in the results directory
        for file in results_dir.glob("*.out"):
            os.remove(file)
    os.chdir(Path.joinpath(test_dir, ".."))
    cmd = "python3 -m pip install -e .[test]"
    subprocess.run(cmd.split(), check=True)
    # Change the current directory to the original directory
    os.chdir(cur_dir)



import copy
import re
from collections import OrderedDict
from io import TextIOWrapper
from typing import List, Tuple
from typing import OrderedDict as ODict

from sum_dirac_dfcoef.atoms import AtomInfo, FuncIndices, ao
from sum_dirac_dfcoef.utils import debug_print, space_separated_parsing


class Function:
    """Data class for storing the information of specific atom + gto_type functions."""

    def __init__(self, component_func: str, symmetry: str, atom: str, gto_type: str, idx_within_same_atom: int, num_functions: int, multiplicity: int) -> None:
        self.component_func = component_func  # "large" or "small"
        self.symmetry = symmetry  # e.g. "Ag"
        self.atom = atom  # e.g. "Cl"
        self.gto_type = gto_type  # e.g. "dxz"
        self.idx_within_same_atom = idx_within_same_atom  # e.g. 1
        self.num_functions = num_functions  # e.g. 3
        self.multiplicity = multiplicity  # e.g. 2


class FunctionsInfo(ODict[str, ODict[str, ODict[str, ODict[int, AtomInfo]]]]):
    """Data class for storing all information about Function numbers and atom labels."""

    # FunctionsInfo(OrderedDict[str, OrderedDict[str, OrderedDict[str, OrderedDict[int, AtomInfo]]]]
    # "large": {
    #     "Ag": {
    #         "Cl": {
    #            "1": {
    #                 AtomInfo: {
    #                     idx_within_same_atom: 1,
    #                     label: "AgCl",
    #                     mul: 2,
    #                     functions: {
    #                         "s": 3,
    #                         "p": 3,
    #                     },
    #                     func_idx_dirac21: {
    #                         first: 1,
    #                         last: 6,
    #                     },
    #                     func_idx_dirac19: {
    #                         first: 1,
    #                         last: 6,
    #                     }
    #                 }
    #            },
    #            "3": {
    #                 AtomInfo: {
    #                     idx_within_same_atom: 3,
    #                     label: "AgCl",
    #                     mul: 2,
    #                     functions: {
    #                         "s": 3,
    #                         "p": 3,
    #                     },
    #                     func_idx_dirac21: {
    #                         first: 7,
    #                         last: 12,
    #                     },
    #                     func_idx_dirac19: {
    #                         first: 7,
    #                         last: 12,
    #                     }
    #                 },...
    #             }
    #         }
    #     }
    # }
    pass


class SymmetryOrbitalsSummary:
    """Class for storing the tuple of the number of orbitals for each symmetry

    Attributes:
        all_sym (Tuple[int]): Number of all orbitals for each symmetry
        large_sym (Tuple[int]): Number of large orbitals for each symmetry
        small_sym (Tuple[int]): Number of small orbitals for each symmetry
    """

    all_sym: Tuple[int]
    large_sym: Tuple[int]
    small_sym: Tuple[int]

    def __init__(self):
        self.all_sym = ()
        self.large_sym = ()
        self.small_sym = ()

    def __repr__(self):
        return f"SymmetryOrbitalsSummary(all_sym: {self.all_sym}, large_sym: {self.large_sym}, small_sym: {self.small_sym})"


class FuncIndicesDIRAC19:
    """Class for storing the current and last function indices for DIRAC19 or older

    Raises:
        ValueError: If the current symmetry index is not the same as the last symmetry index when reading the next symmetry functions.

    Attributes:
        cur_func_idx (int): Current function index
        last_func_idx (int): Last function index
    """

    cur_func_idx: int
    last_func_idx: int

    def __init__(self):
        self.cur_func_idx = 0
        self.last_func_idx = 0

    def __repr__(self):
        return f"FuncIndicesDIRAC19(cur_func_idx: {self.cur_func_idx}, last_func_idx: {self.last_func_idx})"

    def add_num_functions(self, num_functions: int) -> None:
        self.cur_func_idx += num_functions

    def set_indices(self, start: int, last: int) -> None:
        self.cur_func_idx = start
        self.last_func_idx = last

    def check_prev_sym_idx(self) -> None:
        # Before reading the next symmetry, the current symmetry index must be the same as the last symmetry index
        if self.cur_func_idx != self.last_func_idx:
            msg = f"error: The current symmetry index {self.cur_func_idx} must be the same as the last symmetry index {self.last_func_idx}"
            raise ValueError(msg)


class FuncNumDIRAC21:
    """Class for storing the number of functions with gerade, ungerade, and no inversion symmetry

    Attributes:
        gerade (int): Number of functions with gerade symmetry (e.g. "A1g")
        ungerade (int): Number of functions with ungerade symmetry (e.g. "B2u")
        no_inv_sym (int): Number of functions with no inversion symmetry (e.g. "E1")
    """

    gerade: int  # Number of functions with gerade symmetry
    ungerade: int  # Number of functions with ungerade symmetry
    no_inv_sym: int  # Number of functions with no inversion symmetry

    def __init__(self):
        self.gerade = 0
        self.ungerade = 0
        self.no_inv_sym = 0

    def __repr__(self):
        return f"FuncNumDIRAC21(gerade: {self.gerade}, ungerade: {self.ungerade}, no_inv_sym: {self.no_inv_sym})"

    def add_num_functions(self, symmetry: str, num_functions: int) -> None:
        if symmetry[-1].lower() == "g":
            self.gerade += num_functions
        elif symmetry[-1].lower() == "u":
            self.ungerade += num_functions
        else:
            self.no_inv_sym += num_functions

    def get_current_func_idx(self, symmetry: str) -> int:
        if symmetry[-1].lower() == "g":
            return self.gerade
        elif symmetry[-1].lower() == "u":
            return self.ungerade
        else:
            return self.no_inv_sym


class FuncNumSummary:
    """Class for storing the info of the number of function for DIRAC19 and DIRAC21

    Attributes:
        dirac19 (FuncIndicesDIRAC19): The current and last function indices for DIRAC19 or older
        dirac21 (FuncNumDIRAC21): The number of functions with gerade, ungerade, and no inversion symmetry
    """

    dirac19: FuncIndicesDIRAC19
    dirac21: FuncNumDIRAC21

    def __init__(self):
        self.dirac19 = FuncIndicesDIRAC19()
        self.dirac21 = FuncNumDIRAC21()

    def __repr__(self):
        return f"FuncNumSummary(dirac19: {self.dirac19}, dirac21: {self.dirac21})"


def get_functions_info(dirac_output: TextIOWrapper) -> FunctionsInfo:
    def is_start_symmetry_orbitals_section(words: List[str]) -> bool:
        # ref: https://gitlab.com/dirac/dirac/-/blob/de590d17dd38da238ff417b4938d69564158cd7f/src/dirac/dirtra.F#L3654
        if len(words) == 2 and words[0] == "Symmetry" and words[1] == "Orbitals":
            return True
        return False

    def get_component_func(line_str: str) -> str:
        if "Large" in line_str:
            return "large"
        elif "Small" in line_str:
            return "small"
        else:
            msg = f"error: Unknown/Unsupported component function: {line_str}"
            raise ValueError(msg)

    def get_symmetry(words: List[str]) -> str:
        symmetry = words[1]  # e.g. "Ag"
        bra_idx = symmetry.find("(")
        if bra_idx != -1:
            symmetry = symmetry[:bra_idx]
        return symmetry

    def get_symmetry_idx(line_str: str) -> int:
        bra_idx = line_str.find("(")
        ket_idx = line_str.find(")")
        if bra_idx == -1 or ket_idx == -1:
            msg = f"error: The symmetry index is not found in the line: {line_str}"
            raise ValueError(msg)
        return int(line_str[bra_idx + 1 : ket_idx])

    def check_symmetry_idx(idx_symmetry: int, symmetry: str, summary: SymmetryOrbitalsSummary) -> None:
        if idx_symmetry < 0:
            msg = f"error: The symmetry index must be >= 0, but got {idx_symmetry} in symmetry: {symmetry}"
            raise ValueError(msg)
        elif idx_symmetry > len(summary.all_sym):
            msg = f"error: The symmetry index must be <= {len(summary.all_sym)}, but got {idx_symmetry} in symmetry: {symmetry}"
            raise ValueError(msg)

    def read_func_info(words: List[str], line_str: str, component_func: str, symmetry: str) -> Function:
        """Read function information from the external variables line_str and words
        format url: https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/dirac/dirtra.F#L3697-3699
        actual format: 3X,ILAB(1,I)," functions: ",PLABEL(I,2)(6:12),1,(CHRSGN(NINT(CTRAN(II,K))),K,K=2,NDEG)

        (e.g.)  line_str = "18 functions:    Ar s", component_func = "large", symmetry = "Ag"
                => return Function(component_func="large", symmetry="Ag", atom="Ar", gto_type="s", idx_within_same_atom=1, num_functions=18, multiplicity=1)

                line_str = "6 functions:    H  s   1+2", component_func = "large", symmetry = "A1"
                => return Function(component_func="large", symmetry="A1", atom="H", gto_type="s", idx_within_same_atom=1, num_functions=6, multiplicity=2)

        Returns:
            Function: Function information
        """

        def get_last_elem() -> Tuple[int, AtomInfo]:
            # Get the last element of the OrderedDict element with the keys
            last_elem_atom_idx, last_elem = functions_info[component_func][symmetry][atom].popitem()
            # Need to restore the last element because it was popped
            functions_info[component_func][symmetry][atom][last_elem_atom_idx] = last_elem
            return last_elem_atom_idx, last_elem

        def get_idx_within_the_same_atom() -> int:
            try:
                last_elem_atom_idx, last_elem = get_last_elem()
                idx_within_same_atom = last_elem_atom_idx + last_elem.mul
                return idx_within_same_atom
            except KeyError:
                # If the idx_within_same_atom does not exist, it means that this is the first element, so that the idx_within_same_atom is 1
                return 1

        def parse_plabel(plabel: str) -> Tuple[str, str, str]:
            atom = plabel[4:6].strip()  # e.g. "Cm" in "    Cm g400"
            gto_type = plabel[7:].strip()  # e.g. "g400" in "    Cm g400"
            subshell = gto_type[0]  # e.g. "g" in "g400"
            return atom, subshell, gto_type

        def parse_multiplicity_label(multiplicity_label: str) -> int:
            return len(re.findall("[+-]", multiplicity_label)) + 1  # (e.g.) 1+2=>2, 1+2+3=>3, 1+2-3-4=>4

        try:
            num_functions = int(words[0])  # ILAB(1,I) (e.g.) 3
        except ValueError as e:
            # Perhaps words[0] == "******"
            raise ValueError from e  # num_functions must be integer, so raise ValueError and exit this program
        after_functions = line_str[line_str.find("functions:") + len("functions:") :]  # PLABEL(I,2)(6:12),1,(CHRSGN(NINT(CTRAN(II,K))),K,K=2,NDEG) (e.g.) "Cm g400 1+2+3+4
        plabel = after_functions[:11]  # "functions: ",3X,PLABEL(I,2)(6:12) (e.g.) "functions:    Cm g400" => "    Cm g400"
        atom, subshell, gto_type = parse_plabel(plabel)

        multiplicity_label = after_functions[11:].strip()  # 1,(CHRSGN(NINT(CTRAN(II,K))),K,K=2,NDEG) (e.g.) 1+2+3+4
        multiplicity = parse_multiplicity_label(multiplicity_label)  # (e.g.) 4 (1+2+3+4)
        # Set the current subshell and gto_type
        ao.current_ao.update(atom, subshell, gto_type)

        function_label = symmetry + plabel.replace(" ", "")  # symmetry + PLABEL(I,2)(6:12) (e.g.) AgCms
        if ao.is_different_atom(function_label):
            # Different atom
            ao.reset()
            ao.idx_within_same_atom = get_idx_within_the_same_atom()
            # ao was reset, so set the current subshell and gto_type again
            ao.current_ao.update(atom, subshell, gto_type)
            ao.prev_ao = copy.deepcopy(ao.current_ao)

        debug_print(f"function_label: {function_label}, ao: {ao}, idx_within_same_atom: {ao.idx_within_same_atom}")
        ao.function_types.add(function_label)

        return Function(component_func, symmetry, atom, gto_type, ao.idx_within_same_atom, num_functions, multiplicity)

    def update_symmetry_orbitals_summary(line_str: str) -> None:
        tmp_list = [0]  # Add 0 to the beginning of the list
        tmp_list.extend(int(idx) for idx in line_str[line_str.find(":") + 1 :].split())
        indices = tuple(tmp_list)
        if "large" in line_str:
            orb_summary.large_sym = indices
        elif "small" in line_str:
            orb_summary.small_sym = indices
        else:
            orb_summary.all_sym = indices

    def add_function(func: Function) -> None:
        # Create an empty dictionary if the key does not exist
        functions_info.setdefault(component_func, OrderedDict()).setdefault(symmetry, OrderedDict()).setdefault(func.atom, OrderedDict())
        # Add function information if the current atom does not exist in the functions_info (new atom)
        if func.idx_within_same_atom not in functions_info[component_func][symmetry][func.atom].keys():
            label = symmetry + func.atom
            prev_atom_fn_idx_dirac21 = fn_summary.dirac21.get_current_func_idx(symmetry)
            prev_atom_fn_idx_dirac19 = fn_summary.dirac19.cur_func_idx
            fn_idx_dirac21 = FuncIndices(first=prev_atom_fn_idx_dirac21 + 1, last=prev_atom_fn_idx_dirac21 + func.num_functions)
            fn_idx_dirac19 = FuncIndices(first=prev_atom_fn_idx_dirac19 + 1, last=prev_atom_fn_idx_dirac19 + func.num_functions)
            functions_info[component_func][symmetry][func.atom][func.idx_within_same_atom] = AtomInfo(
                func.idx_within_same_atom, label, func.multiplicity, fn_idx_dirac21, fn_idx_dirac19
            )
        functions_info[component_func][symmetry][func.atom][func.idx_within_same_atom].add_function(func.gto_type, func.num_functions)

    def update_last_indices(func: Function) -> None:
        fn_summary.dirac19.add_num_functions(func.num_functions)
        fn_summary.dirac21.add_num_functions(symmetry, func.num_functions)
        functions_info[component_func][symmetry][func.atom][func.idx_within_same_atom].func_idx_dirac19.last = fn_summary.dirac19.cur_func_idx
        functions_info[component_func][symmetry][func.atom][func.idx_within_same_atom].func_idx_dirac21.last = fn_summary.dirac21.get_current_func_idx(symmetry)

    def check_symmetry_orbitals_summary() -> None:
        if len(orb_summary.all_sym) == 0:
            msg = "error: The number of symmetry orbitals is not found."
            raise ValueError(msg)
        elif len(orb_summary.all_sym) != len(orb_summary.large_sym) or len(orb_summary.all_sym) != len(orb_summary.small_sym):
            msg = "error: The number of elements in orb_summary.all_sym, orb_summary.large_sym, and orb_summary.small_sym must be the same."
            raise ValueError(msg)

    start_symmetry_orbitals_section = False
    component_func = "large"  # "large" or "small"
    symmetry = ""
    idx_symmetry = -1
    functions_info = FunctionsInfo()
    fn_summary = FuncNumSummary()
    orb_summary = SymmetryOrbitalsSummary()
    for line_str in dirac_output:
        words: List[str] = space_separated_parsing(line_str)
        if len(line_str) == 0:
            continue
        elif not start_symmetry_orbitals_section:
            start_symmetry_orbitals_section = is_start_symmetry_orbitals_section(words)
        elif "Number of" in line_str:
            update_symmetry_orbitals_summary(line_str)
        elif "component functions" in line_str:
            check_symmetry_orbitals_summary()
            component_func = get_component_func(line_str)
            idx_symmetry = -1
        elif "Symmetry" in line_str:
            fn_summary.dirac19.check_prev_sym_idx()
            symmetry = get_symmetry(words)
            idx_symmetry = get_symmetry_idx(line_str)
            check_symmetry_idx(idx_symmetry, symmetry, orb_summary)
            if component_func == "large":
                start = sum(orb_summary.all_sym[:idx_symmetry])
                last = sum(orb_summary.all_sym[:idx_symmetry]) + orb_summary.large_sym[idx_symmetry]
                fn_summary.dirac19.set_indices(start, last)
            elif component_func == "small":
                start = sum(orb_summary.all_sym[:idx_symmetry]) + orb_summary.large_sym[idx_symmetry]
                last = sum(orb_summary.all_sym[: idx_symmetry + 1])
                fn_summary.dirac19.set_indices(start, last)
        elif "functions" in line_str:
            func = read_func_info(words, line_str, component_func, symmetry)
            add_function(func)
            update_last_indices(func)

        elif all(char in "* \r\n" for char in line_str) and len(re.findall("[*]", line_str)) > 0:
            # all characters in line_str are * or space or line break and at least one * is included
            break  # Stop reading symmetry orbitals
    if not start_symmetry_orbitals_section:
        msg = 'ERROR: The "Symmetry Orbitals" section, which is one of the essential information sections for this program, \
is not in the DIRAC output file.\n\
Please check your DIRAC output file.\n\
Perhaps you explicitly set the .PRINT option to a negative number in one of the sections?'
        raise Exception(msg)

    return functions_info



from typing import ClassVar, List, Tuple


class Subshell:
    """Class for storing all possible subshell labels and their order.

    Attributes:
        orbital_labels (str): All possible orbital labels.
        gto_label_order (List[List[str]]): All possible subshell labels and their order. i.e. gto_label_order = [["s"], ["px", "py", "pz"], ...] after initialization.
    """

    orbital_labels = "spdfghiklmnoqrtuvwxyz"
    gto_label_order: ClassVar[List[List[str]]] = []

    def __lmnval(self, idx: int, nfun: int, istep: List[int], mval: List[int], nval: List[int]) -> Tuple[List[int], List[int], List[int]]:
        # https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/abacus/hergam.F#L218-236
        ix: List[int] = []
        iy: List[int] = []
        iz: List[int] = []
        for i in range(nfun):
            ix.append(idx - istep[i])
            iy.append(mval[i])
            iz.append(nval[i])
        return ix, iy, iz

    def __carpow(self) -> Tuple[List[int], List[int], List[int]]:
        # https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/abacus/hergam.F#L162-216
        istep, mval, nval = [], [], []
        for i in range(1, len(self.orbital_labels) + 1):
            for j in range(1, i + 1):
                istep.append(i)
                mval.append(i - j)
                nval.append(j - 1)
        return istep, mval, nval

    def __init__(self):
        istep, mval, nval = self.__carpow()
        # https://gitlab.com/dirac/dirac/-/blob/b10f505a6f00c29a062f5cad70ca156e72e012d7/src/abacus/herrdn.F#L4995-5021
        for idx, orbital_label in enumerate(self.orbital_labels):
            if orbital_label == "s":
                self.gto_label_order.append(["s"])
            elif orbital_label == "p":
                self.gto_label_order.append(["px", "py", "pz"])
            elif orbital_label == "d":
                self.gto_label_order.append(["dxy", "dxy", "dxz", "dyy", "dyz", "dzz"])
            elif orbital_label == "f":
                self.gto_label_order.append(["fxxx", "fxxy", "fxxz", "fxyy", "fxyz", "fxzz", "fyyy", "fyyz", "fyzz", "fzzz"])
            else:
                nfun = (idx + 1) * (idx + 2) // 2
                ix, iy, iz = self.__lmnval(idx + 1, nfun, istep, mval, nval)
                li = []
                for k in range(nfun):
                    li.append(self.orbital_labels[idx] + str(ix[k]) + str(iy[k]) + str(iz[k]))
                self.gto_label_order.append(li)


subshell_order = Subshell()



from typing import Dict, List, Optional

from sum_dirac_dfcoef.args import args
from sum_dirac_dfcoef.coefficient import Coefficient


class DataMO:
    """This class is used to store the specific MO coefficient information.

    Attributes:
        norm_const_sum (float): The sum of the coefficients of the MO.
        mo_energy (float): The energy of the MO.
        mo_info (str): The information of the MO. (e.g. "Electronic no. 1 E1g")
        sym_type (str): The symmetry type of the MO. (e.g. "E1g")
        eigenvalue_no (int): The eigenvalue number of the MO.
        coef_dict (Dict[str, Coefficient]): The dictionary of the coefficients, in the order written in the DIRAC output file.
        coef_list (List[Coefficient]): The list of the coefficients, which is sorted by the coefficient value.
    """

    norm_const_sum: float = 0.0
    mo_energy: float = 0.0
    mo_info: str = ""
    sym_type: str = ""
    eigenvalue_no: int = 0
    coef_dict: Dict[str, Coefficient]
    coef_list: List[Coefficient]

    def __init__(
        self,
        mo_info: str = "",
        mo_energy: float = 0.0,
        eigenvalue_no: int = 0,
        sym_type: str = "",
        coef_dict: Optional[Dict[str, Coefficient]] = None,
        coef_list: Optional[List[Coefficient]] = None,
    ) -> None:
        self.mo_info = mo_info
        self.mo_energy = mo_energy
        self.eigenvalue_no = eigenvalue_no
        self.sym_type = sym_type
        self.coef_dict = coef_dict if coef_dict is not None else {}
        self.coef_list = coef_list if coef_list is not None else []

    def __repr__(self) -> str:
        return f"DataMO(mo_info: {self.mo_info}, mo_energy: {self.mo_energy}, eigenvalue_no: {self.eigenvalue_no}, mo_sym_type: {self.sym_type}, coef_dict: {self.coef_dict})"

    def add_coefficient(self, coef: Coefficient) -> None:
        def generate_key(coef: Coefficient) -> str:
            key = (coef.atom_label, coef.azimuthal_label, coef.idx_within_same_atom)
            if not args.ignore_sym:
                key += (coef.symmetry_label,)
            if not args.ignore_ml:
                key += (coef.magnetic_label,)
            return key

        key = generate_key(coef)
        if key in self.coef_dict:
            self.coef_dict[key].coefficient += coef.coefficient
        else:
            self.coef_dict[key] = coef
        self.norm_const_sum += coef.coefficient * coef.multiplication

    def reset(self):
        self.norm_const_sum = 0.0
        self.mo_energy = 0.0
        self.mo_info = ""
        self.eigenvalue_no = 0
        self.coef_dict.clear()
        self.coef_list.clear()

    def fileter_coefficients_by_threshold(self) -> None:
        self.coef_list = [coef for coef in self.coef_dict.values() if abs(coef.coefficient / self.norm_const_sum * 100) >= args.threshold]
        self.coef_list.sort(key=lambda coef: coef.coefficient, reverse=True)


class DataAllMO:
    """This class stores all electronic and positronic MO data.

    This class holds information used for output excluding header information.

    Attributes:
        electronic (List[DataMO]): The list of the electronic MO data.
        positronic (List[DataMO]): The list of the positronic MO data.
    """

    electronic: List[DataMO]
    positronic: List[DataMO]

    def __init__(self, electronic: Optional[List[DataMO]] = None, positronic: Optional[List[DataMO]] = None) -> None:
        self.electronic = electronic if electronic is not None else []
        self.positronic = positronic if positronic is not None else []

    def __repr__(self) -> str:
        return f"DataAllMO(electronic: {self.electronic}, positronic: {self.positronic})"

    def sort_mo_sym_type(self) -> None:
        self.electronic.sort(key=lambda mo: (mo.sym_type, mo.mo_energy))
        self.positronic.sort(key=lambda mo: (mo.sym_type, mo.mo_energy))

    def sort_mo_energy(self) -> None:
        self.electronic.sort(key=lambda mo: mo.mo_energy)
        self.positronic.sort(key=lambda mo: mo.mo_energy)



from collections import OrderedDict
from typing import OrderedDict as ODict
from typing import Set

from pydantic import BaseModel, validator

from sum_dirac_dfcoef.subshell import subshell_order


class FuncIndices:
    """Class for storing the first and last function indices of the specified atom label(AtomInfo).

    Attributes:
        first (int): The first function index of the specified atom label.
        last (int): The last function index of the specified atom label.
    """

    first: int
    last: int

    def __init__(self, first: int = 0, last: int = 0) -> None:
        self.first = first
        self.last = last

    def __repr__(self) -> str:
        return f"FunIndices(first: {self.first}, last: {self.last})"


class AtomInfo:
    """Class for handling the information of the specified atom label.

    This class is used for FunctionInfo class and PrivecProcessor class.

    Attributes:
        idx_within_same_atom (int): Index of the order of atoms in the same AtomLabel
        label (str): The label of the specified atom label.
        mul (int): The multiplicity of the specified atom label.
        functions (ODict[str, int]): The number of functions of the specified atom label.
        func_idx_dirac21 (FuncIndices): The first and last function indices of the specified atom label. (For DIRAC >= 21)
        func_idx_dirac19 (FuncIndices): The first and last function indices of the specified atom label. (For DIRAC < 21)
    """

    idx_within_same_atom: int
    label: str
    mul: int
    functions: ODict[str, int]
    func_idx_dirac21: FuncIndices
    func_idx_dirac19: FuncIndices  # For DIRAC < 21

    def __init__(
        self,
        idx_within_same_atom: int = 0,
        label: str = "",
        multiplicity: int = 0,
        func_idx_dirac21: FuncIndices = None,
        func_idx_dirac19: FuncIndices = None,
    ) -> None:
        self.idx_within_same_atom = idx_within_same_atom
        self.label = label
        self.mul = multiplicity
        self.functions = OrderedDict()
        self.func_idx_dirac21 = func_idx_dirac21 if func_idx_dirac21 is not None else FuncIndices()
        self.func_idx_dirac19 = func_idx_dirac19 if func_idx_dirac19 is not None else FuncIndices()

    def __repr__(self) -> str:
        return f"AtomInfo(idx_within_same_atom {self.idx_within_same_atom}, label: {self.label}, mul: {self.mul}, \
func_idx_dirac21: {self.func_idx_dirac21}, func_idx_dirac19: {self.func_idx_dirac19}, functions: {self.functions})"

    def add_function(self, gto_type: str, num_functions: int) -> None:
        self.functions[gto_type] = num_functions

    def decrement_function(self, gto_type: str) -> None:
        try:
            self.functions[gto_type] -= 1
            if self.functions[gto_type] < 0:
                msg = f"Too many functions detected. self.functions[{gto_type}] must be >= 0, but got {self.functions[gto_type]}"
                raise ValueError(msg)
        except KeyError as e:
            msg = f"self.functions[{gto_type}] is not found in self.functions: {self.functions.keys()}"
            raise KeyError(msg) from e

    def get_remaining_functions(self) -> "ODict[str, int]":
        return OrderedDict({k: v for k, v in self.functions.items() if v > 0})


class AtomicOrbital(BaseModel, validate_assignment=True):
    """Class for handling the information of the atomic orbital.

    Raises:
        ValueError: Raises an error if the subshell is not in the subshell_order.
        ValueError: Raises an error if the subshell is not one character.

    Attributes:
        atom (str): The label of the atom. (e.g. "C", "H", "He")
        subshell (str): The label of the subshell. (e.g. "s", "p", "d", "f")
        gto_type (str): The label of the GTO type. (e.g. "s", "px", "dxy", "fxxx")
    """

    atom: str = ""
    subshell: str = "s"
    gto_type: str = "s"

    @validator("subshell")
    def validate_subshell(cls, v: str) -> str:  # noqa: N805 (pydantic method)
        if v not in subshell_order.orbital_labels:
            msg = f"subshell must be one of '{subshell_order.orbital_labels}', but got '{v}'"
            raise ValueError(msg)
        if len(v) != 1:
            msg = f"subshell must be one character, but got '{v}'"
            raise ValueError(msg)
        return v

    def reset(self):
        self.atom = ""
        self.subshell = "s"
        self.gto_type = "s"

    def update(self, atom: str, subshell: str, gto_type: str):
        self.atom = atom
        self.subshell = subshell
        self.gto_type = gto_type


class AtomicOrbitals(BaseModel, validate_assignment=True):
    """Class for handling the information of the atomic orbitals.

    Main purpose of this class is to check if the current ao and the previous ao are the same atom.

    Attributes:
        prev_ao (AtomicOrbital): The previous atomic orbital.
        current_ao (AtomicOrbital): The current atomic orbital.
        idx_within_same_atom (int): The index of the order of atoms in the same AtomLabel.
        function_types (Set[str]): The set of function types.

    Methods:
        reset: Reset the AtomicOrbitals instance.
        is_different_atom: Check if the current ao and the previous ao are the same atom.
    """

    prev_ao: AtomicOrbital = AtomicOrbital()
    current_ao: AtomicOrbital = AtomicOrbital()
    idx_within_same_atom: int = 1
    function_types: Set[str] = set()

    def reset(self):
        self.prev_ao.reset()
        self.current_ao.reset()
        self.idx_within_same_atom = 1
        self.function_types.clear()

    def is_different_atom(self, function_label: str) -> bool:
        def is_reverse_subshell() -> bool:
            prev_subshell_idx = subshell_order.orbital_labels.index(self.prev_ao.subshell)
            current_subshell_idx = subshell_order.orbital_labels.index(self.current_ao.subshell)
            if prev_subshell_idx > current_subshell_idx:
                return True
            elif prev_subshell_idx < current_subshell_idx:
                return False
            else:  # Same subshell
                subshell_idx = subshell_order.orbital_labels.index(self.prev_ao.subshell)
                prev_gto_idx = subshell_order.gto_label_order[subshell_idx].index(self.prev_ao.gto_type)
                current_gto_idx = subshell_order.gto_label_order[subshell_idx].index(self.current_ao.gto_type)
                if prev_gto_idx > current_gto_idx:  # reverse subshell. e.g. self.prev_ato.gto_type = "pz", self.current_ao.gto_type = "px"
                    return True
                else:
                    return False

        if self.prev_ao.atom != self.current_ao.atom:
            return True
        elif function_label in self.function_types:
            # They have the same element label but different atoms.
            # Because the function_label of an atom is combined in one line,
            # it is a different atom if the same function_label appears.
            return True
        elif is_reverse_subshell():
            # e.g. "C  p" -> "C  s"
            # Different atom
            return True
        return False  # Same atom


ao = AtomicOrbitals()
